import pysam
import pandas as pd

import handygenome.workflow as workflow
import handygenome.variant.varianthandler as varianthandler
import handygenome.signature.misc as signature_misc
import handygenome.signature.sigprofiler as sigprofiler
import handygenome.signature.cataloguing as cataloguing
import handygenome.signature.plotter_sbs96 as plotter_sbs96
import handygenome.variant.vcfspec as libvcfspec
import handygenome.deco as deco


LOGGER = workflow.get_logger(__name__, level='info')


class SignatureResult:
    def __init__(self, 
                 catalogue: pd.Series, 
                 sigdata: pd.DataFrame, 
                 exposure: pd.Series, 
                 cossim,
                 kldiv=None, correlation=None):
        def sanity_check(catalogue, sigdata, exposure):
            if tuple(catalogue.index) != tuple(sigdata.index):
                raise Exception(
                    f'Index of "catalogue" and index of "sigdata" '
                    f'are different')
            if tuple(exposure.index) != tuple(sigdata.columns):
                raise Exception(
                    f'Index of "exposure" and columns of "sigdata" '
                    f'are different')

        sanity_check(catalogue, sigdata, exposure)

        self.catalogue = catalogue
        self.sigdata = sigdata
        self.exposure = exposure
        self.cossim = cossim
        self.kldiv = kldiv
        self.correlation = correlation

        self.catalogue_type = signature_misc.get_catalogue_type(self.catalogue.index)
        self.nonzero_sigs = tuple(
            self.exposure.index[self.exposure > 0])

    def plot(self, sampleid=None):
        if self.catalogue_type == 'sbs96':
            plotter_sbs96.main(self, sampleid=sampleid)
        else:
            raise Exception(f'Unavailable catalogue type')


@deco.get_deco_arg_choices({'refver': signature_misc.AVAILABLE_REFVERS})
@deco.get_deco_arg_choices({'catalogue_type': ('sbs96', 'id83')})
@deco.get_deco_arg_choices({'cataloguer': ('custom', 'sigprofiler')})
def get_sigresult_from_vcfspecs(vcfspec_iter, refver='GRCh37', 
                                catalogue_type='sbs96',
                                cataloguer='custom',
                                verbose=True,
                                verbose_matgen=True,
                                verbose_assign=False, 
                                background_sigs=list(), 
                                permanent_sigs=list(),
                                checkrule_sigs=list(), 
                                add_penalty=0.05, 
                                remove_penalty=0.01,
                                checkrule_penalty=1.00,
                                use_recommended_sigs=True,
                                use_connected_sigs=True):
    # get catalogue
    if verbose:
        LOGGER.info('Creating mutation catalogue')

    if cataloguer == 'custom':
        if catalogue_type == 'sbs96':
            catalogue = cataloguing.get_sbs96_catalogue_vcfspecs(vcfspec_iter, 
                                                                 refver)
        else:
            raise Exception(f'Custom cataloguer is available only for sbs96.')
    elif cataloguer == 'sigprofiler':
        catalogues = sigprofiler.get_catalogues_from_vcfspecs(
            vcfspec_iter, refver, verbose=verbose_matgen)
        catalogue = catalogues[catalogue_type].iloc[:, 0]

    # loading signature database
    sigdata = signature_misc.load_signature_data(refver, catalogue_type)

    # fitting signature components
    if verbose:
        LOGGER.info('Running SigProfilerAssignment')

    sigresult = sigprofiler.run_assignment(
        catalogue, sigdata,
        background_sigs=background_sigs, 
        permanent_sigs=permanent_sigs,
        checkrule_sigs=checkrule_sigs, 
        add_penalty=add_penalty, 
        remove_penalty=remove_penalty,
        checkrule_penalty=checkrule_penalty,
        use_recommended_sigs=use_recommended_sigs,
        use_connected_sigs=use_connected_sigs,
        verbose=verbose_assign)

    if verbose:
        LOGGER.info('All finished.')

    return sigresult


@deco.get_deco_arg_choices({'refver': signature_misc.AVAILABLE_REFVERS})
def get_sigresult_from_vcfpath(vcf_path, refver='GRCh37', **kwargs):
    vcfspec_iter = (
        libvcfspec.Vcfspec.from_vr(vr)
        #varianthandler.get_vcfspec(vr)
        for vr in pysam.VariantFile(vcf_path).fetch()
    )
    sigresult = get_sigresult_from_vcfspecs(vcfspec_iter, refver, **kwargs)

    return sigresult




import os
import itertools
import functools

import Bio.SeqUtils
import pyBigWig
import numpy as np
import pandas as pd
import pyranges as pr

import handygenome.common as common


GCWIG_DIR = os.path.join(common.DATA_DIR, 'gcwig')
GCARRAY_DIR = os.path.join(common.DATA_DIR, 'gcarrays')


# making and saving gc fraction bigwig files

def write_gcwig(outfile_path, fasta, binsize=100, verbose=False):
    bw = pyBigWig.open(outfile_path, 'w')
    # make header
    header = list() 
    for chrom, length in zip(fasta.references, fasta.lengths):
        header.append((chrom, length))
    bw.addHeader(header)
    # write lines
    for chrom, length in zip(fasta.references, fasta.lengths):
        if verbose:
            print(chrom)
        rng = range(0, length, binsize)
        start0s = (x for x in rng)
        end0s = (min(length, x + binsize) for x in rng)
        chroms = itertools.repeat(chrom, len(rng))
        values = calculate_gcvals(chroms, start0s, end0s, fasta, window=None, as_array=False)
        bw.addEntries(chrom, 0, values=values, span=binsize, step=binsize)

    bw.close()


def get_gcwig_path(refver, binsize):
    refver = common.RefverDict.standardize(refver)
    gcwig_dir = os.path.join(GCWIG_DIR, refver)
    if not os.path.exists(gcwig_dir):
        os.mkdir(gcwig_dir)
    return os.path.join(gcwig_dir, f'gc_binsize{100}.bigwig')


def get_gcwig(refver, binsize):
    gcfile_path = get_gcwig_path(refver, binsize)
    if not os.path.exists(gcfile_path):
        common.print_timestamp(f'There is no pre-existing gc bigwig file. A new one is being created. It may take a few minutes.')
        fasta = common.DEFAULT_FASTAS[refver]
        write_gcwig(gcfile_path, fasta, binsize=binsize, verbose=False)
        common.print_timestamp(f'Finished making a gc bigwig file.')
    return pyBigWig.open(gcfile_path)


###########################################

# getting array of gc fraction values

def make_fetchargs_func(window, fasta):
    if window is None:
        def get_fetchargs(chrom, start0, end0):
            return chrom, start0, end0
    else:
        chromlens = dict(zip(fasta.references, fasta.lengths))
        pad_left = int(window / 2)
        if window % 2 == 0:
            pad_right = pad_left
        else:
            pad_right = pad_left + 1

        def get_fetchargs(chrom, start0, end0):
            if end0 - start0 >= window:
                return chrom, start0, end0
            else:
                mid = int((start0 + end0) / 2)
                new_start0 = max(0, mid - pad_left)
                new_end0 = min(chromlens[chrom], mid + pad_right)
                return chrom, new_start0, new_end0

    return get_fetchargs


def calculate_gcvals_generator(chroms, start0s, end0s, fasta, get_fetchargs):
    prev_chrom = ''
    for chrom, start0, end0 in zip(chroms, start0s, end0s):
        if chrom != prev_chrom:
            print(chrom)
        prev_chrom = chrom
        yield float(Bio.SeqUtils.gc_fraction(fasta.fetch(*get_fetchargs(chrom, start0, end0))))


@functools.cache
def calculate_gcvals(chroms, start0s, end0s, fasta, window=None, as_array=True):
    print('Beginning gc fraction calculation')
    get_fetchargs = make_fetchargs_func(window, fasta)
    #val_gen = calculate_gcvals_generator(chroms, start0s, end0s, fasta, get_fetchargs)
    val_gen = (
        float(Bio.SeqUtils.gc_fraction(fasta.fetch(*get_fetchargs(chrom, start0, end0))))
        for chrom, start0, end0 in zip(chroms, start0s, end0s)
    )

    if as_array:
        return np.fromiter(val_gen, float)
    else:
        return list(val_gen)


def calculate_gcvals_with_df(df, fasta, window=None, as_array=True):
    return calculate_gcvals(
        tuple(df.Chromosome), 
        tuple(df.Start), 
        tuple(df.End), 
        fasta, 
        window=window, 
        as_array=as_array,
    )


# MUCH SLOWER THAN calculate_gcvals...
def load_gcvals(chroms, start0s, end0s, refver, binsize, fasta, window=None, as_array=True, exact=True):
    get_fetchargs = make_fetchargs_func(window, fasta)
    gcwig = get_gcwig(refver, binsize)
    val_gen = (
        gcwig.stats(
            *get_fetchargs(chrom, start0, end0), 
            type='mean', 
            exact=exact,
        )[0]
        for chrom, start0, end0 in zip(chroms, start0s, end0s)
    )

    if as_array:
        return np.fromiter(val_gen, float)
    else:
        return list(val_gen)


def load_gcvals_with_df(df, refver, binsize, fasta, window=None, as_array=True, exact=True):
    return load_gcvals(
        df.Chromosome, df.Start, df.End, 
        refver=refver, binsize=binsize, 
        fasta=common.DEFAULT_FASTAS[refver], 
        window=window, as_array=as_array,
        exact=exact,
    )


# add gc fraction values to a depth dataframe

@common.get_deco_num_set_differently(('fasta', 'refver'), 1)
def add_gc_calculating(df, *, refver=None, fasta=None, window=None):
    """Args:
        df: pandas.DataFrame or pyranges.PyRanges
        refver, fasta: refver is recommended because with refver, default 
            fasta object with the given refver is loaded and subsequently
            used for "calculate_gcvals" function. Using a fasta object with
            a fixed id can make use of "cache" functionality.

    Changes in-place
    """
    if fasta is None:
        fasta = common.DEFAULT_FASTAS[refver]

    gcvals = calculate_gcvals(
        tuple(df.Chromosome), 
        tuple(df.Start), 
        tuple(df.End), 
        fasta, 
        window=window, 
        as_array=True,
    )

    if isinstance(df, pd.DataFrame):
        df['GC'] = gcvals
    elif isinstance(df, pr.PyRanges):
        df.GC = gcvals


def add_gc_loading(df, refver, binsize, window=None):
    """Args:
        df: pandas.DataFrame or pyranges.PyRanges
    Changes in-place
    """
    gcvals = load_gcvals(
        df.Chromosome, df.Start, df.End, 
        refver=refver, binsize=binsize, 
        fasta=common.DEFAULT_FASTAS[refver], 
        window=window, as_array=False,
    )
    if isinstance(df, pd.DataFrame):
        df['GC'] = gcvals
    elif isinstance(df, pr.PyRanges):
        df.GC = gcvals





import os
import re
import argparse
import gzip

import pysam

import handygenome.tools as tools
import handygenome.network as network
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.refgenome.refgenome as refgenome


DEFAULT_OUTFILE_BASENAME = 'ucsc_repeatmasker_out_sorted.bed.gz'
URL_GRCH37 = 'https://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/hg19.fa.out.gz'
URL_GRCH38 = 'https://hgdownload.soe.ucsc.edu/goldenPath/hg38/bigZips/hg38.fa.out.gz'
DEFAULT_OUTFILE_PATH_GRCH37 = os.path.join(os.getcwd(), DEFAULT_OUTFILE_BASENAME)
DEFAULT_OUTFILE_PATH_GRCH38 = os.path.join(os.getcwd(), DEFAULT_OUTFILE_BASENAME)


def argument_parser(cmdargs):
    parser_dict = workflow.init_parser(
        description=(f'The original RepeatMasker output files are '
                     f'downloaded from ucsc, then filtered, sorted, '
                     f'bgzipped, and tabix-indexed.'))

    parser_dict['optional'].add_argument(
        '--outfile-grch37', dest='outfile_path_grch37', 
        default=DEFAULT_OUTFILE_PATH_GRCH37,
        required=False, help=f'Output repeat bed file path for grch37')
    parser_dict['optional'].add_argument(
        '--outfile-grch38', dest='outfile_path_grch38', 
        default=DEFAULT_OUTFILE_PATH_GRCH38,
        required=False, help=f'Output repeat bed file path for grch38')
    parser_dict['optional'].add_argument(
        '--downloaded-file-grch37', dest='download_path_grch37', 
        required=False, 
        help=(f'(Optional) previously downloaded original ucsc file path '
              f'for grch37'))
    parser_dict['optional'].add_argument(
        '--downloaded-file-grch38', dest='download_path_grch38', 
        required=False, 
        help=(f'(Optional) previously downloaded original ucsc file path '
              f'for grch38'))

    workflow.add_logging_args(parser_dict)

    args = parser_dict['main'].parse_args(cmdargs)

    return args


def download(logger, url, download_path):
    logger.info(f'Beginning download of url "{url}" to "{download_path}"')
    network.download(url, download_path)
    logger.info(f'Download finished')


def parse(download_path, asmblspec, output_chrname_version):
    parse_result = list()
    with gzip.open(download_path, 'rt') as infile:
        for _ in range(3):
            line = next(infile)

        for line in infile:
            linestrip = tools.rm_newline(line)
            linesp = linestrip.split()

            chrom = asmblspec.convert(linesp[4], output_chrname_version)
            start = int(linesp[5]) - 1
            end = int(linesp[6])
            name = linesp[10] + ':' + linesp[9]
            score = int(linesp[0])
            if linesp[8] == '+':
                strand = '+'
            elif linesp[8] == 'C':
                strand = '-'
            else:
                raise Exception(f'Unexpected strand string of repeatmasker '
                                f'output: {linesp[8]}')

            line_parse = [chrom, start, end, name, score, strand]
            parse_result.append(line_parse)

    return parse_result


def sort_parse_result(chromdict, parse_result):
    def sortkey(line_parse):
        return tools.coord_sortkey(line_parse[0], line_parse[1], chromdict)
    parse_result_sorted = sorted(parse_result, key=sortkey)

    return parse_result_sorted


def write_uncompressed_file(outfile_path_uncomp, parse_result_sorted):
    with open(outfile_path_uncomp, 'wt') as outfile:
        outfile.write(
            '\t'.join(['#contig', 'start', 'end', 'class/family:name', 
                       'SW_score', 'strand']) + '\n')
        for line_parse in parse_result_sorted:
            outfile.write('\t'.join(str(x) for x in line_parse) + '\n')


def convert_common(original_url, output_chrname_version, refver, logger, 
                   outfile_path, download_path=None):
    # download
    if download_path is None:
        download_path = workflow.get_tmpfile_path()
        download(logger, original_url, download_path)
        rm_downloaded = True
    else:
        rm_downloaded = False

    # set refgenome dbs
    asmblspec = refgenome.SPECS[refver]
    chromdict = asmblspec.chromdicts[output_chrname_version]

    # load
    logger.info(f'Loading the original file')
    parse_result = parse(download_path, asmblspec, output_chrname_version)

    # sort
    logger.info(f'Sorting the original file lines')
    parse_result_sorted = sort_parse_result(chromdict, parse_result)

    # write uncompressed outfile
    logger.info(f'Writing the sorted lines')
    outfile_path_uncomp = outfile_path + '.uncompressed'
    write_uncompressed_file(outfile_path_uncomp, parse_result_sorted)

    # compress and index
    logger.info(f'Compressing with bgzip and indexing with tabix')
    pysam.tabix_compress(outfile_path_uncomp, outfile_path)
    pysam.tabix_index(outfile_path, preset='bed')

    # remove temp files
    os.remove(outfile_path_uncomp)
    if rm_downloaded:
        os.remove(download_path)


def convert_grch37(logger, outfile_path, download_path=None):
    convert_common(URL_GRCH37, 'nochr_plus_genbank', 'hg19', logger, 
                   outfile_path, download_path)


def convert_grch38(logger, outfile_path, download_path=None):
    convert_common(URL_GRCH38, 'ucsc', 'hg38', logger, outfile_path, 
                   download_path)


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    convert_grch37(logger, args.outfile_path_grch37, args.download_path_grch37)
    convert_grch38(logger, args.outfile_path_grch38, args.download_path_grch38)
    
    logger.info('ALL SUCCESSFULLY FINISHED')



import re
import itertools
import collections
import functools
import inspect
import logging
import uuid

import numpy as np
import pandas as pd

import handygenome.tools as tools
import handygenome.refgenome.refgenome as refgenome
import handygenome.workflow as workflow
import handygenome.read.readhandler as readhandler
import handygenome.align.alignhandler as alignhandler
import handygenome.read.readplus as readplus


class PileupBase:
    pat_insclip = re.compile("(\(.+\))?([^()]+)(\(.+\))?")
    pat_parenthesis = re.compile("[()]")
    DEL_VALUE = "*"
    EMPTY_VALUE = "_"
    DF_FILL_VALUES = {
        'main': EMPTY_VALUE,
        'bq': -1,
        'queryonly': '',
        'queryonly_bq': np.nan,
        'onleft': False,
        'unedited_main': EMPTY_VALUE,
        'queryonly_added_main': EMPTY_VALUE,
    }
        
    ##########################
    # methods for direct use #
    ##########################
    @property
    def start0(self):
        return self.seq_df.columns[0]

    @property
    def end0(self):
        return self.seq_df.columns[-1] + 1

    @property
    def range0(self):
        return range(self.start0, self.end0)

    @property
    def width(self):
        return self.seq_df.shape[1]

    def get_ref_seq(self):
        return self.fasta.fetch(self.chrom, self.start0, self.end0)

    def get_depth(self, pos0):
        col = self.seq_df.loc[:, pos0]
        return (col != self.__class__.EMPTY_VALUE).sum()
        #return sum(x != self.__class__.EMPTY_VALUE for x in col)

    def get_queryonly_added_df(self):
        main_asnp = self.dfs['main'].to_numpy().astype(str)
        queryonly_asnp = self.dfs['queryonly'].to_numpy().astype(str)

        queryonly_exists = (queryonly_asnp != self.__class__.DF_FILL_VALUES['queryonly'])
        main_replaced = np.where(
            (main_asnp == self.__class__.DF_FILL_VALUES['main']),
            '',
            main_asnp,
        )
        result = np.where(
            queryonly_exists,
            np.where(
                self.dfs['onleft'], 
                np.char.add(queryonly_asnp, main_asnp),
                np.char.add(main_asnp, queryonly_asnp),
            ),
            main_asnp,
        )
        result = pd.DataFrame(
            result, 
            index=self.dfs['main'].index, 
            columns=self.dfs['main'].columns, 
        )
        return result

    def get_allele_counter(self, pos0):
        col = self.dfs['queryonly_added_main'].loc[:, pos0]
        return collections.Counter(col[col != self.__class__.EMPTY_VALUE])
        #return collections.Counter(x for x in col if x != self.__class__.EMPTY_VALUE)

    def get_allele_portions(self, pos0):
        counter = self.get_allele_counter(pos0)
        counter_sum = sum(counter.values())
        portions = dict()
        for key, val in counter.items():
            portions[key] = val / counter_sum
        return portions

    def get_read(self, read_uid):
        return self.read_store[read_uid]

    def get_lowbq_replaced_df(self, main_df, bq_df, cutoff=15):
        ref_seq = self.get_ref_seq()

        bq_exists = (bq_df != self.__class__.DF_FILL_VALUES['bq'])
        bq_is_low = np.logical_and(
            bq_exists,
            bq_df < cutoff,
        )
        bq_is_low = bq_is_low.to_numpy()

        result = main_df.copy()
        for row_idx, col_idx in zip(*bq_is_low.nonzero()):
            result.iloc[row_idx, col_idx] = ref_seq[col_idx]

        return result

    @property
    def seq_df(self):
        try:
            return self.dfs['queryonly_added_main']
        except KeyError:
            return self.dfs['main']
            

    ############
    # backends #
    ############
    def __init__(
        self, 
        bam, chrom, 
        start0=None, end0=None, 
        refver=None, fasta=None,

        init_df=True, 

        lowbq_cutoff=15,

        verbose=False, 
        logger=None,
    ):
        """Args:
            start0 and end0 are only required when init_df == True
        """
        # set params
        self.bam = bam
        self.chrom = chrom
        self.refver, self.fasta = self.refver_fasta_arghandler(refver, fasta)

        # set logger
        self.verbose, self.logger = self.logger_arghandler(verbose, logger)

        # init df and subsequent ones
        if init_df:
            (
                df, bq_df, queryonly_df, queryonly_bq_df, onleft_df, read_store,
            )= make_pileup_components(
                chrom,
                start0,
                end0,
                bam,
                truncate=True,
                as_array=False,
                verbose=self.verbose,
                logger=self.logger,
            )
            self.dfs = {
                'main': df,
                'bq': bq_df,
                'queryonly': queryonly_df,
                'queryonly_bq': queryonly_bq_df,
                'onleft': onleft_df,
            }
            self.read_store = read_store

            # replace lowbq with REF
            lowbq_replaced_df = self.get_lowbq_replaced_df(
                main_df=self.dfs['main'],
                bq_df=self.dfs['bq'],
                cutoff=lowbq_cutoff,
            )
            self.dfs['unedited_main'] = self.dfs['main']
            self.dfs['main'] = lowbq_replaced_df

            # make queryonly-added main
            self.dfs['queryonly_added_main'] = self.get_queryonly_added_df()

            self._set_MQ()
        else:
            self.dfs = None
            self.read_store = None
            self.MQ = None

    def __repr__(self):
        return (
            f'<'
            f'{self.__class__.__name__}'
            f'(chrom={self.chrom}, start0={self.start0:,}, end0={self.end0:,}, shape={self.seq_df.shape})'
            f'>'
        )

    def _coord_arg_sanitycheck(self, start0, end0):
        #if start0 is not None:
        if start0 < self.start0:
            raise Exception('Input "start0" argument is out of pileup range.')
        #if end0 is not None:
        if end0 > self.end0:
            raise Exception('Input "end0" argument is out of pileup range.')

    def _coord_arg_sanitycheck_pos0(self, pos0):
        if pos0 not in self.seq_df.columns:
            raise Exception('Input "pos0" argument is out of pileup range.')

    def _set_MQ(self):
        if self.seq_df.shape[1] == 0:
            self.MQ = pd.Series([], dtype=float)
        else:
            if self.seq_df.shape[0] > 0:
                MQ_data = {pos0: list() for pos0 in self.seq_df.columns}
                #for read_uid, row in self.df.iterrows():
                for read_uid in self.seq_df.index:
                    read = self.read_store[read_uid]
                    start0 = max(self.start0, read.reference_start)
                    end0 = min(self.end0, read.reference_end)
                    for pos0 in range(start0, end0):
                        MQ_data[pos0].append(read.mapping_quality)
                self.MQ = pd.Series(
                    [
                        np.nan if (len(MQ_data[pos0]) == 0) else np.mean(MQ_data[pos0])
                        for pos0 in self.seq_df.columns
                    ],
                    index=self.seq_df.columns,
                )
            else:
                self.MQ = pd.Series(
                    [np.inf] * self.seq_df.shape[1],
                    index=self.seq_df.columns,
                )

    def _spawn_base(self, attrs_to_copy, init_df=False):
        """Make a partial copy of self with essential attributes"""
        #for key in inspect.signature(self.__init__).parameters.keys():
        #    if key not in ('init_df', 'start0', 'end0'):
        #        kwargs[key] = getattr(self, key)

        kwargs = dict()
        for key in attrs_to_copy: 
            kwargs[key] = getattr(self, key)
        kwargs['init_df'] = init_df

        result = self.__class__(**kwargs)
        result.read_store = self.read_store
        return result

    def _split_base(self, start0_list):
        assert all((x in range(self.start0 + 1, self.end0)) for x in start0_list)
        result = list()
        margins = [self.start0] + start0_list + [self.end0]
        for new_start0, new_end0 in tools.pairwise(margins):
            result.append(self.subset(new_start0, new_end0, inplace=False))
        return result

    def _subset_base(self, start0, end0, inplace=False):
        #subset_df, subset_bq_df = self._get_subset_df(start0, end0)
        subset_dfs = self._get_subset_df(start0, end0)
        new_MQ = self.MQ.loc[start0:(end0 - 1)]
        if inplace:
            self.dfs = subset_dfs
            self.MQ = new_MQ
            return None
        else:
            result = self.spawn()
            result.dfs = subset_dfs
            result.MQ = new_MQ
            return result

    def _get_subset_df(self, start0, end0):
        self._coord_arg_sanitycheck(start0, end0)
        # get row selector
        row_within_range = list()
        for row_id in self.seq_df.index:
            read = self.read_store[row_id]
            if read.reference_end <= start0 or read.reference_start >= end0:
                row_within_range.append(False)
            else:
                row_within_range.append(True)

        # subset dataframe
        new_dfs = {
            key: val.loc[row_within_range, start0:(end0 - 1)]
            for key, val in self.dfs.items()
        }
        #new_df = self.df.loc[row_within_range, start0:(end0 - 1)]
        #new_bq_df = self.bq_df.loc[row_within_range, start0:(end0 - 1)]

        return new_dfs

    def _get_subset_df_old(self, start0, end0):
        self._coord_arg_sanitycheck(start0, end0)
        # subset dataframe
        new_df = self.df.loc[:, start0:(end0 - 1)]
        new_bq_df = self.bq_df.loc[:, start0:(end0 - 1)]
        # remove out-of-range rows
        row_within_range = list()
        #for row_id, row in new_df.iterrows():
        for row_id in new_df.index:
            read = self.read_store[row_id]
            if read.reference_end <= start0 or read.reference_start >= end0:
                row_within_range.append(False)
            else:
                row_within_range.append(True)
        new_df = new_df.loc[row_within_range, :]
        new_bq_df = new_bq_df.loc[row_within_range, :]

        return new_df, new_bq_df

    def _merge_base(self, other, other_on_left):
        if other_on_left:
            left = other
            right = self
        else:
            left = self
            right = other

        assert left.end0 == right.start0
        border_col_idx_left = left.seq_df.shape[1] - 1
        new_dfs = {
            key: left.dfs[key].join(right.dfs[key], how="outer")
            for key in left.dfs.keys()
        }
        #for key, df in zip(keys, vals):
        for key in list(new_dfs.keys()):
            df = new_dfs[key]
            try:
                df.where(df.notna(), other=self.__class__.DF_FILL_VALUES[key], inplace=True)
            except TypeError as exc:
                if str(exc) == 'Cannot do inplace boolean setting on mixed-types with a non np.nan value':
                    new_dfs[key] = df.where(
                        df.notna(), other=self.__class__.DF_FILL_VALUES[key], inplace=False,
                    )
                else:
                    raise
                    
        self.dfs = new_dfs
        #self.df = left.df.join(right.df, how="outer")
        #self.bq_df = left.bq_df.join(right.bq_df, how="outer")
        self.MQ = pd.concat([left.MQ, right.MQ])

#        try:
#            self.df[self.df.isnull()] = self.__class__.EMPTY_VALUE  # turn NaN into EMPTY_VALUE
#        except:
#            print(self.df.shape)
#            print(self.df)
#            print(type(self.df.iloc[0, 0]))
#            raise
        #self._handle_facing_insclips(border_col_idx_left)
        self.read_store.update(other.read_store)

    def _handle_facing_insclips(self, border_col_idx_left):
        border_col_idx_right = border_col_idx_left + 1
        border_col_left = self.df.iloc[:, border_col_idx_left]
        border_col_right = self.df.iloc[:, border_col_idx_right]
        for row_idx, (val_left, val_right) in enumerate(
            zip(border_col_left, border_col_right)
        ):
            if val_right.startswith("(") and val_left.endswith(")"):
                mat_left = self.__class__.pat_insclip.fullmatch(val_left)
                mat_right = self.__class__.pat_insclip.fullmatch(val_right)
                if mat_left.group(3) != mat_right.group(1):
                    raise Exception(
                        f"Insclip seqs of adjacent entries are different.\n{self.df}"
                    )
                self.df.iloc[
                    row_idx, border_col_idx_right
                ] = self.__class__.pat_insclip.sub("\\2\\3", val_right)

    def _extend_base(self, width, left):
        new_pileup = self._make_extend_pileup(width, left)
        self.merge(new_pileup, other_on_left=left)

    @staticmethod
    def refver_fasta_arghandler(refver, fasta):
        if (refver is None) and (fasta is None):
            raise Exception(f'At least one of "refver" or "fasta" must be set.')

        if refver is None:
            refver_result = refgenome.infer_refver_fasta(fasta)
        else:
            refver_result = refver

        if fasta is None:
            fasta_result = refgenome.get_fasta(refver_result)
        else:
            fasta_result = fasta

        return refver_result, fasta_result

    @classmethod
    def logger_arghandler(cls, verbose, logger):
        verbose_result = verbose
        if logger is None:
            logger_result = cls.get_logger(verbose_result)
        else:
            logger_result = logger

        return verbose_result, logger_result

    @staticmethod
    def get_logger(verbose):
        formatter = logging.Formatter(
            fmt='[%(asctime)s.%(msecs)03d] Pileup: %(message)s', 
            datefmt='%Z %Y-%m-%d %H:%M:%S'
        )
        return workflow.get_logger(
            name=str(uuid.uuid4()),
            level=('debug' if verbose else 'info'),
            formatter=formatter,
        )


class Pileup(PileupBase):
    # initializers
    def __init__(
        self, 
        bam, chrom, 
        start0=None, end0=None, 
        refver=None, fasta=None,
        init_df=True, 
        verbose=False, logger=None,
    ):
        PileupBase.__init__(
            self, 
            bam=bam, chrom=chrom, 
            start0=start0, end0=end0, 
            refver=refver, fasta=fasta,
            init_df=init_df, 
            verbose=verbose, logger=logger,
        )

    def subset(self, start0, end0, inplace=False):
        return self._subset_base(self, start0, end0, inplace=inplace)

    def merge(self, other, other_on_left):
        self._merge_base(other, other_on_left=other_on_left)

    def split(self, start0_list):
        return self._split_base(start0_list)

    def spawn(self):
        return self._spawn_base(
            ('bam', 'chrom', 'refver', 'fasta', 'verbose', 'logger'),
            init_df=False,
        )

    # extend
    def extend_left(self, width):
        self._extend_base(width, left=True)

    def extend_right(self, width):
        self._extend_base(width, left=False)

    def _make_extend_pileup(self, width, left):
        if left:
            start0 = self.start0 - width
            end0 = self.start0
        else:
            start0 = self.end0
            end0 = self.end0 + width
        return self.__class__(
            fasta=self.fasta, 
            bam=self.bam, 
            chrom=self.chrom, 
            start0=start0,
            end0=end0,
            init_df=True,
        )


class MultisamplePileup(PileupBase):
    def __init__(self, pileup_dict):
        """Args:
            pileup_dict: keys - sampleid; values - Pileup object
        """
        self.pileup_dict = pileup_dict

        self.first_pileup = next(iter(self.pileup_dict.values()))
        self.chrom = self.first_pileup.chrom
        self.fasta = self.first_pileup.fasta

    def _get_active_threshold(self):
        depth_sum = sum(x.df.shape[0] for x in self.pileup_dict.values())
        corrected_active_thresholds = (
            x.active_threshold * (x.df.shape[0] / depth_sum)
            for x in self.pileup_dict.values()
        )
        return min(corrected_active_thresholds)

    def _get_active_threshold_simple(self):
        return sum(x.active_threshold for x in self.pileup_dict.values()) / (len(self.pileup_dict) ** 2)

    @property
    def start0(self):
        return self.first_pileup.start0

    @property
    def end0(self):
        return self.first_pileup.end0

    def set_df(self):
        assert len(set(x.range0 for x in self.pileup_dict.values())) == 1, f'Genomic ranges of pileup objects are different.'
        self.df = pd.concat(
            {key: val.df for key, val in self.pileup_dict.items()},
            names=['SampleID', 'ReadUID'],
        )
        #self._set_active_info()

    def get_read(self, row_id):
        sampleid, read_uid = row_id
        return self.pileup_dict[sampleid].read_store[read_uid]

    ### extend ###
    def extend_rightward(self, width):
        for pileup in self.pileup_dict.values():
            pileup.extend_rightward(width)

    def extend_leftward(self, width):
        for pileup in self.pileup_dict.values():
            pileup.extend_leftward(width)

    # equalize
    def equalize_margins_leftward(self):
        max_range_start0 = min(x.start0 for x in self.pileup_dict.values())
        for pileup in self.pileup_dict.values():
            if pileup.start0 > max_range_start0:
                pileup.extend_leftward(pileup.start0 - max_range_start0)

    def equalize_margins_rightward(self):
        max_range_end0 = max(x.end0 for x in self.pileup_dict.values())
        for pileup in self.pileup_dict.values():
            if pileup.end0 < max_range_end0:
                pileup.extend_rightward(max_range_end0 - pileup.end0)

    def equalize_margins(self):
        self.equalize_margins_leftward()
        self.equalize_margins_rightward()

    ### row specs ###
#    def set_row_specs(self, set_for_each=False):
#        if set_for_each:
#            for pileup in self.pileup_dict.values():
#                pileup.set_row_specs()
#
#        self.row_specs = dict()
#        for row_id, row in self.df.iterrows():
#            sampleid, read_uid = row_id
#            self.row_specs[row_id] = self.pileup_dict[sampleid].row_specs[read_uid]

    def set_row_spec_groups(self):
        """Assumes row_spec_groups is set for each single sample Pileup"""
        # make superseq_row_specs
        candidate_superseq_row_specs = dict()
        for sampleid, pileup in self.pileup_dict.items():
            for read_uid in pileup.row_spec_groups.keys():
                candidate_superseq_row_specs[(sampleid, read_uid)] = pileup.row_specs[read_uid]
        superseq_row_specs = dict()
        for superseq_rowid in self.__class__.group_row_specs(candidate_superseq_row_specs).keys():
            superseq_row_specs[superseq_rowid] = candidate_superseq_row_specs[superseq_rowid]
        # make row_spec_groups for each sample
        row_spec_groups_by_sample = dict()
        for sampleid, pileup in self.pileup_dict.items():
            row_spec_groups = collections.OrderedDict()
            for superseq_rowid in superseq_row_specs.keys():
                row_spec_groups[superseq_rowid] = {
                    'subseq_rowids': set(),
                    'subseq_hits': 0,
                }
            for query_read_uid, query_row_spec in pileup.row_specs.items():
                superseq_candidates = list()
                for superseq_rowid, superseq_row_spec in superseq_row_specs.items():
                    if self.__class__.row_spec_matcher(query_row_spec, superseq_row_spec):
                        superseq_candidates.append(superseq_rowid)
                    #if len(superseq_candidates) >= 2:
                    #    break

                if len(superseq_candidates) == 0:
                    raise Exception(f'row_spec in sub-pileup does not fit to any of superseqs:\nsampleid: {sampleid}\nrow_spec of subseq: {query_row_spec}')
                else:
                    self.__class__.handle_matching_subseq(superseq_candidates, row_spec_groups, query_read_uid)

            row_spec_groups_by_sample[sampleid] = row_spec_groups
        # final    
        self.superseq_row_specs = superseq_row_specs
        self.row_spec_groups = row_spec_groups_by_sample

    def save_superseq_alignments(self):
        librealign.save_superseq_alignments_multisample(self)

    # get realigned reads #
    def get_realigned_reads(self):
        """Must be run by those created from 'get_active_region_pileup_multisample' function

        Returns:
            dict (keys sampleid, values dict (keys ReadUID, values pysam.AlignedSegment))
        """
        return librealign.get_realigned_reads_multisample(self)

    def set_vcfspecs(self, allele_portion_threshold=None, concat_dist_le=None):
        """Must be run by those created from 'get_active_region_pileup_multisample' function"""
        self.vcfspecs = librealign.get_vcfspecs_from_pileup_multisample(self, allele_portion_threshold=allele_portion_threshold, concat_dist_le=concat_dist_le)

    # for debugging
    def show_row_spec_alignments(self, skip_zero_hits=True, show_subseqs=True):
        for sampleid, pileup in self.pileup_dict.items():
            print('@@@', sampleid, '@@@')
            print()
            self._show_row_spec_alignments_helper(self.row_spec_groups[sampleid], self.superseq_row_specs, skip_zero_hits, show_subseqs)


def make_pileup_components(
    chrom,
    start0,
    end0,
    bam,
    truncate=True,
    as_array=False,
    del_value=PileupBase.DEL_VALUE,
    empty_value=PileupBase.EMPTY_VALUE,
    verbose=False,
    logger=None,
    trailing_queryonly_to_right=False,
    preserve_trailing_queryonly=True,
):
    """Args:
        trailing_queryonly_to_right: 
            - If True, insertions or softclips on the right border of the read
                are assigned to "read.reference_end"
            - If False, assigned to "read.reference_end - 1"
        preserve_trailing_queryonly:
            - Relevant only when "truncate" is True and "trailing_queryonly_to_right" is True.
            - If True, if trailing queryonly is placed on "end0" position,
                returned array or dataframe gets to include the position "end0"
                to accommodate the trailing queryonly sequence. In this case,
                returned "ref_span_range" becomes range(start0, end0 + 1).
            - If False, if trailing queryonly is placed on "end0" position,
                they are not included in the returned array or dataframe,
                and "ref_span_range" becomes range(start0, end0).
    """
    # set logger
    if logger is None:
        logger = PileupBase.get_logger(verbose)
    else:
        logger = logger

    def make_read_store(bam, chrom, start0, end0):
        target_range0 = range(start0, end0)
        readlist = list()
        uid_list = list()
        start_list = list()
        end_list = list()
        for read in readhandler.get_fetch(bam, chrom, start0, end0, readfilter=readhandler.readfilter_pileup):
            if readhandler.check_cigarN_includes_range(read, start0, end0):
                continue

#            target_seq = readplus.ReadPlus(read, minimal=True).get_seq_from_range0(
#                target_range0,
#                flanking_queryonly_default_mode=True,
#            )
#            if target_seq == '':
#                print(read)
#                continue

            readlist.append(read)
            uid_list.append(readhandler.get_uid(read))
            start_list.append(read.reference_start)
            end_list.append(read.reference_end)

        if len(start_list) == 0:
            tmp_pileup_range = None
        else:
            tmp_pileup_range = range(
                min(min(start_list), start0), 
                max(max(end_list), end0) + 1,
            )

        read_store = collections.OrderedDict(zip(uid_list, readlist))

        return read_store, tmp_pileup_range

    def raise_cigarpattern_error(read):
        raise Exception(f"Unexpected cigar pattern:\n{read.to_string()}")

    def cigar_sanitycheck(read):
        """Assumes M.* or [IS]M.*"""
        error = False
        first_cigarop = read.cigartuples[0][0]
        if first_cigarop != 0:
            if first_cigarop not in (1, 4):
                error = True
            else:
                if read.cigartuples[1][0] != 0:
                    error = True
        if error:
            raise_cigarpattern_error(read)

    def truncate_array(arr, tmp_pileup_range, ref_span_range):
        sl_start = tmp_pileup_range.index(ref_span_range.start)
        if ref_span_range.stop == tmp_pileup_range.stop:
            sl_end = None
        else:
            sl_end = tmp_pileup_range.index(ref_span_range.stop)
        return arr[:, slice(sl_start, sl_end)]

    def make_array(
        start0, end0, tmp_pileup_range, read_store, empty_value, del_value, trailing_queryonly_to_right,
    ):
        if tmp_pileup_range is None:  # zero depth
            ref_span_range = range(start0, end0)
            arr = np.empty((0, len(ref_span_range)))
            bq_arr = arr.copy()
            queryonly_arr = arr.copy()
            queryonly_bq_arr = arr.copy()
            onleft_arr = arr.copy()
        else:
            # initialize array
            arr = np.full(
                shape=(len(read_store), len(tmp_pileup_range)),
                fill_value=PileupBase.DF_FILL_VALUES['main'],
                dtype=str,
            )
            bq_arr = np.full(
                shape=(len(read_store), len(tmp_pileup_range)),
                fill_value=PileupBase.DF_FILL_VALUES['bq'],
                dtype=int,
            )
            queryonly_arr = np.full(
                shape=(len(read_store), len(tmp_pileup_range)),
                fill_value=PileupBase.DF_FILL_VALUES['queryonly'],
                dtype=object,
            )
            queryonly_bq_arr = np.full(
                shape=(len(read_store), len(tmp_pileup_range)),
                fill_value=PileupBase.DF_FILL_VALUES['queryonly_bq'],
                dtype=object,
            )
            onleft_arr = np.full(
                shape=(len(read_store), len(tmp_pileup_range)),
                fill_value=PileupBase.DF_FILL_VALUES['onleft'],
                dtype=bool,
            )

            for arr_row_idx, read in enumerate(read_store.values()):
                arr_col_idx = read.reference_start - tmp_pileup_range.start
                try:
                    _write_read_to_array_row(
                        read, arr, arr_row_idx, arr_col_idx, del_value, 
                        bq_arr, queryonly_arr, queryonly_bq_arr, onleft_arr, 
                        trailing_queryonly_to_right,
                    )
                except Exception as exc:
                    try:
                        cigar_sanitycheck(read)
                    except Exception as exc_cigarpattern:
                        raise exc_cigarpattern from exc
                    else:
                        raise exc

            last_col_is_empty = (set(arr[:, -1]) == {empty_value})
            if last_col_is_empty:
                arr = arr[:, :-1]
                bq_arr = bq_arr[:, :-1]
                queryonly_arr = queryonly_arr[:, :-1]
                queryonly_bq_arr = queryonly_bq_arr[:, :-1]
                onleft_arr = onleft_arr[:, :-1]
                tmp_pileup_range = range(tmp_pileup_range.start, tmp_pileup_range.stop - 1)

            # truncate array
            if truncate:
                ref_span_range = range(start0, end0)
                arr = truncate_array(arr, tmp_pileup_range, ref_span_range)
                bq_arr = truncate_array(bq_arr, tmp_pileup_range, ref_span_range)
                queryonly_arr = truncate_array(queryonly_arr, tmp_pileup_range, ref_span_range)
                queryonly_bq_arr = truncate_array(queryonly_bq_arr, tmp_pileup_range, ref_span_range)
                onleft_arr = truncate_array(onleft_arr, tmp_pileup_range, ref_span_range)
            else:
                ref_span_range = tmp_pileup_range

        return arr, bq_arr, queryonly_arr, queryonly_bq_arr, onleft_arr, ref_span_range

    # main
    read_store, tmp_pileup_range = make_read_store(bam, chrom, start0, end0)
        # tmp_pileup_range contains end0 of the rightmost fetched read

    arr, bq_arr, queryonly_arr, queryonly_bq_arr, onleft_arr, ref_span_range = make_array(
        start0, end0, tmp_pileup_range, read_store, empty_value, del_value, trailing_queryonly_to_right,
    )

    # final result
    if as_array:
        return arr, bq_arr, queryonly_arr, queryonly_bq_arr, onleft_arr, ref_span_range, read_store
    else:
        df = pd.DataFrame(
            arr, 
            columns=list(ref_span_range), 
            index=pd.Index(read_store.keys(), tupleize_cols=False),
        )
        bq_df = pd.DataFrame(
            bq_arr, 
            columns=list(ref_span_range), 
            index=pd.Index(read_store.keys(), tupleize_cols=False),
        )
        queryonly_df = pd.DataFrame(
            queryonly_arr, 
            columns=list(ref_span_range), 
            index=pd.Index(read_store.keys(), tupleize_cols=False),
        )
        queryonly_bq_df = pd.DataFrame(
            queryonly_bq_arr, 
            columns=list(ref_span_range), 
            index=pd.Index(read_store.keys(), tupleize_cols=False),
        )
        onleft_df = pd.DataFrame(
            onleft_arr, 
            columns=list(ref_span_range), 
            index=pd.Index(read_store.keys(), tupleize_cols=False),
        )
        return df, bq_df, queryonly_df, queryonly_bq_df, onleft_df, read_store


def _write_read_to_array_row(
    read, arr, arr_row_idx, arr_col_idx, del_value, 
    bq_arr, queryonly_arr, queryonly_bq_arr, onleft_arr, 
    trailing_queryonly_to_right=True,
    bq_array_includes_queryonly=False,
):
    """Helper function for 'make_pileup_components'"""
    query_idx = 0
    queryonly_seq = None
    for idx, (key, subiter) in enumerate(
        itertools.groupby(
            read.cigartuples, 
            key=(lambda x: alignhandler.CIGAR_WALK_DICT[x[0]]),
        )
    ):
        consume_target, consume_query = key
        cigarlen_sum = sum(x[1] for x in subiter)

        if consume_target and consume_query:  # match
            if queryonly_seq is None:
                sl_query = slice(query_idx, query_idx + cigarlen_sum)
                sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen_sum)
                arr[arr_row_idx, sl_arr_col] = tuple(read.query_sequence[sl_query])

                bq_arr[arr_row_idx, sl_arr_col] = tuple(read.query_qualities[sl_query])  # BQ array
            else:
                # add query-only information
                arr[arr_row_idx, arr_col_idx] = read.query_sequence[query_idx]
                bq_arr[arr_row_idx, arr_col_idx] = read.query_qualities[query_idx]
                queryonly_arr[arr_row_idx, arr_col_idx] = queryonly_seq
                queryonly_bq_arr[arr_row_idx, arr_col_idx] = queryonly_bq
                onleft_arr[arr_row_idx, arr_col_idx] = True

                # add trailing matched bases
                sl_query = slice(query_idx + 1, query_idx + cigarlen_sum)
                sl_arr_col = slice(arr_col_idx + 1, arr_col_idx + cigarlen_sum)
                arr[arr_row_idx, sl_arr_col] = tuple(read.query_sequence[sl_query])
                bq_arr[arr_row_idx, sl_arr_col] = tuple(read.query_qualities[sl_query])  # BQ array

                queryonly_seq = None
                queryonly_bq = None

            arr_col_idx += cigarlen_sum
            query_idx += cigarlen_sum

        elif (not consume_target) and consume_query:  # ins, softclip
            queryonly_seq = read.query_sequence[query_idx:(query_idx + cigarlen_sum)]
            queryonly_bq = tuple(read.query_qualities[query_idx:(query_idx + cigarlen_sum)])  # BQ array
            query_idx += cigarlen_sum

        elif consume_target and (not consume_query):  # del, skip
            if queryonly_seq is not None:
                raise Exception(f'Cigar D or N comes right after I or S. Current read: {read.to_string()}')

            sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen_sum)
            arr[arr_row_idx, sl_arr_col] = del_value

            arr_col_idx += cigarlen_sum

    # When the last cigar operation is query-only
    if queryonly_seq is not None:
        if idx == 0:  # query-only-only case
            queryonly_arr[arr_row_idx, arr_col_idx] = queryonly_seq
            queryonly_bq_arr[arr_row_idx, arr_col_idx] = queryonly_bq  # BQ array
        else:  # queryonly_seq is appended to the base on the left
            queryonly_arr[arr_row_idx, arr_col_idx - 1] = queryonly_seq
            onleft_arr[arr_row_idx, arr_col_idx - 1] = False
            queryonly_bq_arr[arr_row_idx, arr_col_idx - 1] = queryonly_bq


def _write_read_to_array_row_old(
    read, arr, arr_row_idx, arr_col_idx, del_value, 
    bq_arr, queryonly_arr, onleft_arr, 
    trailing_queryonly_to_right=True,
    bq_array_includes_queryonly=False,
):
    """Helper function for 'make_pileup_components'"""
    query_idx = 0
    queryonly_seq = None
    for idx, (key, subiter) in enumerate(
        itertools.groupby(
            read.cigartuples, 
            key=(lambda x: alignhandler.CIGAR_WALK_DICT[x[0]]),
        )
    ):
        consume_target, consume_query = key
        cigarlen_sum = sum(x[1] for x in subiter)

        if consume_target and consume_query:  # match
            if queryonly_seq is None:
                sl_query = slice(query_idx, query_idx + cigarlen_sum)
                sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen_sum)
                arr[arr_row_idx, sl_arr_col] = tuple(read.query_sequence[sl_query])

                bq_arr[arr_row_idx, sl_arr_col] = tuple(read.query_qualities[sl_query])  # BQ array
            else:
                # add the first match-base, prefixed with query-only sequence buffer
                arr[arr_row_idx, arr_col_idx] = f'({queryonly_seq})' + read.query_sequence[query_idx]
                if bq_array_includes_queryonly:  # BQ array
                    bq_arr[arr_row_idx, arr_col_idx] = (
                        queryonly_bq, 
                        read.query_qualities[query_idx],
                    )  
                else:
                    bq_arr[arr_row_idx, arr_col_idx] = read.query_qualities[query_idx]

                # add trailing matched bases
                sl_query = slice(query_idx + 1, query_idx + cigarlen_sum)
                sl_arr_col = slice(arr_col_idx + 1, arr_col_idx + cigarlen_sum)
                arr[arr_row_idx, sl_arr_col] = tuple(read.query_sequence[sl_query])
                bq_arr[arr_row_idx, sl_arr_col] = tuple(read.query_qualities[sl_query])  # BQ array

                queryonly_seq = None
                queryonly_bq = None

            arr_col_idx += cigarlen_sum
            query_idx += cigarlen_sum

        elif (not consume_target) and consume_query:  # ins, softclip
            queryonly_seq = read.query_sequence[query_idx:(query_idx + cigarlen_sum)]
            queryonly_bq = tuple(read.query_qualities[query_idx:(query_idx + cigarlen_sum)])  # BQ array
            query_idx += cigarlen_sum

        elif consume_target and (not consume_query):  # del, skip
            if queryonly_seq is not None:
                raise Exception(f'Cigar D or N comes right after I or S. Current read: {read.to_string()}')

            sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen_sum)
            arr[arr_row_idx, sl_arr_col] = del_value
            bq_arr[arr_row_idx, sl_arr_col] = del_value  # BQ array

            arr_col_idx += cigarlen_sum

    # When the last cigar operation is query-only
    if queryonly_seq is not None:
        if idx == 0:  # query-only-only case
            arr[arr_row_idx, arr_col_idx] = f'({queryonly_seq})'
            if bq_array_includes_queryonly:  # BQ array
                bq_arr[arr_row_idx, arr_col_idx] = queryonly_bq  # BQ array
        else:  # queryonly_seq is appended to the base on the left
            if trailing_queryonly_to_right:
                arr[arr_row_idx, arr_col_idx] = f'({queryonly_seq})'
                if bq_array_includes_queryonly:  # BQ array
                    bq_arr[arr_row_idx, arr_col_idx] = queryonly_bq  # BQ array
            else:
                arr[arr_row_idx, arr_col_idx - 1] = arr[arr_row_idx, arr_col_idx - 1] + f'({queryonly_seq})'
                if bq_array_includes_queryonly:  # BQ array
                    bq_arr[arr_row_idx, arr_col_idx - 1] = (
                        bq_arr[arr_row_idx, arr_col_idx - 1],
                        queryonly_bq,
                    )  # BQ array


def _write_read_to_array_row_deprecated(read, arr, arr_row_idx, initial_arr_col_idx, del_value):
    """NOT USED. THIS VERSION ASSIGNS INSCLIP TO THE POSITION ON THE LEFT."""
    def handle_leading_insclip(read):
        leading_insclip_len = 0
        cigartup_idx = 0
        for cigarop, cigarlen in alignhandler.iter_leading_queryonly(read.cigartuples):
            leading_insclip_len += cigarlen
            cigartup_idx += 1

        if leading_insclip_len == 0:
            leading_insseq = None
        else:
            leading_insseq = read.query_sequence[:leading_insclip_len]
        query_idx = leading_insclip_len

        return leading_insseq, cigartup_idx, query_idx

    def handle_queryonly_only_case(read, arr, arr_col_idx, arr_row_idx):
        arr[arr_row_idx, arr_col_idx] = "(" + read.query_sequence + ")"            

    def handle_targetonly_after_match(
        read,
        arr,
        query_idx,
        arr_col_idx,
        arr_row_idx,
        trailing_nonM_cigarlen_sum,
        del_value,
    ):
        arr[arr_row_idx, arr_col_idx] = read.query_sequence[
            query_idx
        ]  # query_idx indicates the last base of M
        arr[
            arr_row_idx,
            (arr_col_idx + 1) : (arr_col_idx + 1 + trailing_nonM_cigarlen_sum),
        ] = del_value

    def handle_queryonly_after_match(
        read,
        arr,
        query_idx,
        arr_col_idx,
        arr_row_idx,
        trailing_nonM_cigarlen_sum,
    ):
        last_match_base = read.query_sequence[query_idx]
        insseq = read.query_sequence[
            (query_idx + 1) : (query_idx + 1 + trailing_nonM_cigarlen_sum)
        ]
        arr[arr_row_idx, arr_col_idx] = f"{last_match_base}({insseq})"

    def handle_nonlast_match(
        read,
        arr,
        cigarlen,
        query_idx,
        arr_col_idx,
        arr_row_idx,
        cigartup_idx,
    ):
        # adds all match seqs except the last one
        # query_idx and arr_col_idx updated
        if cigarlen > 1:
            sl_query = slice(query_idx, query_idx + cigarlen - 1)
            sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen - 1)
            arr[arr_row_idx, sl_arr_col] = tuple(read.query_sequence[sl_query])
            query_idx += cigarlen - 1
            arr_col_idx += cigarlen - 1
        # get all subsequent non-M cigar units
        # cigartup_idx updated
        trailing_nonMs = list()
        while True:
            cigartup_idx += 1
            if cigartup_idx >= len(read.cigartuples):
                break
            else:
                next_cigartup = read.cigartuples[cigartup_idx]
                if next_cigartup[0] != 0:
                    trailing_nonMs.append(next_cigartup)
                else:
                    break
        # add trailing non-M cigar units
        # query_idx and arr_col_idx updated
        trailing_nonM_cigarops = set(x[0] for x in trailing_nonMs)
        trailing_nonM_cigarlen_sum = sum(x[1] for x in trailing_nonMs)
        if trailing_nonM_cigarops.issubset(alignhandler.CIGAROPS_TARGETONLY):  # D, N
            handle_targetonly_after_match(
                read,
                arr,
                query_idx,
                arr_col_idx,
                arr_row_idx,
                trailing_nonM_cigarlen_sum,
                del_value,
            )
            query_idx += 1
            arr_col_idx += trailing_nonM_cigarlen_sum + 1
        elif trailing_nonM_cigarops.issubset(alignhandler.CIGAROPS_QUERYONLY):  # I, S
            handle_queryonly_after_match(
                read,
                arr,
                query_idx,
                arr_col_idx,
                arr_row_idx,
                trailing_nonM_cigarlen_sum,
            )
            query_idx += trailing_nonM_cigarlen_sum + 1
            arr_col_idx += 1
        else:
            raise Exception(f"Consecutive Non-M cigar units are composed of both target-only and query-only ones:\n{read.to_string()}")

        return cigartup_idx, query_idx, arr_col_idx

    def handle_last_match(
        read,
        arr,
        cigarlen,
        query_idx,
        arr_col_idx,
        arr_row_idx,
    ):
        # adds all match seqs to array
        sl_query = slice(query_idx, query_idx + cigarlen)
        sl_arr_col = slice(arr_col_idx, arr_col_idx + cigarlen)
        arr[arr_row_idx, sl_arr_col] = tuple(read.query_sequence[sl_query])

    # main
    """1) First, leading query-only cigar units are collected.
    2) Then, cigartuple is iterated on. Initial cigartup_idx is at the 
        first cigar unit which is not a query-only one. This cigar unit
        is assumed to be M. (If not, raises an exception)
    3) What is done in a loop cycle:
        The cigar unit which cigartup_idx indicates (which should be M)
        and all subsequent non-M cigarop are treated at single loop cycle.
        The subsequent non-M cigar units are assumed to be composed of only
        query-only ones or target-only ones. (If not, raises an exception)
    """

    # set mutable arr_col_idx. initial one is kept separately.
    arr_col_idx = initial_arr_col_idx

    # handles leading query-only cigar units. cigartup_idx and query_idx are initialized
    leading_insseq, cigartup_idx, query_idx = handle_leading_insclip(read)

    if cigartup_idx == len(read.cigartuples):
        # cigartup is only composed of query-only operations (I or S)
        handle_queryonly_only_case(read, arr, arr_col_idx, arr_row_idx)
    else:
        if read.cigartuples[cigartup_idx][0] != 0:
            raise Exception(
                f"The first cigar operation after stripping leading I and S is not M:\n{read.to_string()}"
            )

        # begins loop
        while True:
            if cigartup_idx > len(read.cigartuples) - 1:
                raise Exception(f'"cigartup_idx" became greater than "len(read.cigartuples) - 1" while looping.')

            cigarop, cigarlen = read.cigartuples[cigartup_idx]
            if cigarop != 0:
                raise Exception(f'The cigar unit indicated by "cigarop_idx" at the beginning of a loop cycle is not M.')

            if cigartup_idx == len(read.cigartuples) - 1:  
                # current cigartup is the last one
                handle_last_match(
                    read,
                    arr,
                    cigarlen,
                    query_idx,
                    arr_col_idx,
                    arr_row_idx,
                )
                break
            else:
                cigartup_idx, query_idx, arr_col_idx = handle_nonlast_match(
                    read,
                    arr,
                    cigarlen,
                    query_idx,
                    arr_col_idx,
                    arr_row_idx,
                    cigartup_idx,
                )
                if cigartup_idx == len(read.cigartuples):
                    break

        # add leading query-only seq
        if leading_insseq is not None:
            arr[arr_row_idx, initial_arr_col_idx] = (
                f"({leading_insseq})" + arr[arr_row_idx, initial_arr_col_idx]
            )


#def get_pileup(
#    chrom,
#    start0,
#    end0,
#    bam,
#    fasta,
#    truncate=True,
#    del_value=DEL_VALUE,
#    empty_value=EMPTY_VALUE,
#):
#    df, read_store = make_pileup_components(
#        chrom,
#        start0,
#        end0,
#        bam,
#        truncate=truncate,
#        as_array=False,
#        del_value=del_value,
#        empty_value=empty_value,
#    )
#    pileup = Pileup(df, chrom, fasta, bam, read_store)
#
#    return pileup





#def get_pileup_multisample(
#    chrom,
#    start0,
#    end0,
#    bam_dict,
#    fasta,
#    active_threshold_onesample=None,
#    del_value=PileupBase.DEL_VALUE,
#    empty_value=PileupBase.EMPTY_VALUE,
#):
#    # parameter handling
#    if active_threshold_onesample is None:
#        active_threshold_onesample = librealign.DEFAULT_ACTIVE_THRESHOLD
#    # initialize pileup_dict
#    pileup_dict = dict()
#    for sampleid, bam in bam_dict.items():
#        pileup_dict[sampleid] = get_pileup(
#            chrom,
#            start0,
#            end0,
#            bam=bam,
#            fasta=fasta,
#            active_threshold=active_threshold_onesample,
#            truncate=True,
#            as_array=False,
#            return_range=False,
#            del_value=del_value,
#            empty_value=empty_value,
#        )
#    # create MultisamplePileup object and postprocess
#    mspileup = MultisamplePileup(pileup_dict)
#    mspileup.set_df()
#
#    return mspileup






import os
import re
import argparse
import gzip
import textwrap

import pysam
import pyranges as pr

import importlib
top_package_name = __name__.split('.')[0]
common = importlib.import_module('.'.join([top_package_name, 'common']))
workflow = importlib.import_module('.'.join([top_package_name, 'workflow']))
sequenza = importlib.import_module('.'.join([top_package_name, 'cnv', 'sequenza']))


PROGNAME = 'seqz_maker'


def unit_job(tbam_path, nbam_path, regionfile_path, split_outfile_subdir,
             refver, fasta_path, gcwiggle_path, hom, het):
    intvlist = common.IntervalList.from_gr(pr.read_bed(regionfile_path))
    for zidx, intv in common.zenumerate(intvlist):
        split_outfile_path = os.path.join(split_outfile_subdir, 
                                          f'{zidx}.seqz.gz')
        sequenza.seqzutils_bam2seqz(
            tbam_path=tbam_path, nbam_path=nbam_path, 
            outfile_path=split_outfile_path, 
            refver=refver, fasta_path=fasta_path, 
            gcwiggle_path=gcwiggle_path, 
            force_outfile_name=False,
            chrom=intv.chrom, start0=intv.start0, end0=intv.end0,
            hom=hom, het=het)


def argument_parser(cmdargs):
    def sanity_check(args):
        sequenza.bam2seqz_sanity_check(
            args.force_outfile_name, args.outfile_path, args.refver, 
            args.fasta_path, args.gcwiggle_path)

    parser_dict = workflow.init_parser(
        description=(
            f'1) Whole genome is split into N intervals; '
            f'2) sequenza-utils bam2seqz is run for each interval; '
            f'3) Split results are concatenated; '
            f'4) sequenza-utils seqz_binning is done.\n\n'
            f'Preset fasta and gcwiggle files are used when --refver option '
            f'is used.\n'
            f'When --refver is set, --fasta and --gcwiggle must not be set.\n'
            f'When --refver is not set, --fasta and --gcwiggle must be set.\n'
            ))
    # required
    workflow.add_outfile_arg(parser_dict['required'], required=True, 
                             must_not_exist='ask')
    parser_dict['required'].add_argument(
        '-t', '--tbam', dest='tbam_path', required=True,
        help=f'Tumor bam file path.')
    parser_dict['required'].add_argument(
        '-n', '--nbam', dest='nbam_path', required=True,
        help=f'Normal bam file path.')

    # optional
    workflow.add_refver_arg(parser_dict['optional'], required=False, 
                            choices=('hg19', 'hg38'))
    workflow.add_fasta_arg(parser_dict['optional'], required=False)

    workflow.add_incl_region_arg(parser_dict['optional'], required=False)
    workflow.add_excl_region_arg(parser_dict['optional'], required=False)

    workflow.add_logging_args(parser_dict)
    workflow.add_scheduler_args(parser_dict, default_parallel=1, 
                                default_sched='slurm')

    parser_dict['optional'].add_argument(
        '--gcwiggle', dest='gcwiggle', required=False,
        help=f'GC wiggle file path.')
    parser_dict['optional'].add_argument(
        '--hom-threshold', dest='hom', required=False,
        type=float, default=sequenza.DEFAULT_HOM,
        help=(f'--hom option value to be used in sequenza-utils bam2seqz. '
              f'Represents VAF threshold to treat as homozygous SNP.'))
    parser_dict['optional'].add_argument(
        '--het-threshold', dest='het', required=False,
        type=float, default=sequenza.DEFAULT_HET,
        help=(f'--het option value to be used in sequenza-utils bam2seqz. '
              f'Represents VAF threshold to treat as heterozygous SNP.'))

    # flag
    workflow.add_rmtmp_arg(parser_dict)
    parser_dict['flag'].add_argument(
        '--force-outfile-name', dest='force_outfile_name', 
        action='store_true',
        help=f'When not set, output file name must end with ".seqz.gz".')

    args = parser_dict['main'].parse_args(cmdargs)
    return args


def make_tmpdir(infile_path):
    tmpdir_paths = workflow.get_tmpdir_paths(
        ['scripts', 'logs', 'regions', 'split_outfiles'],
        prefix=(os.path.basename(infile_path) + '_' + PROGNAME),
        where=os.path.dirname(infile_path))

    return tmpdir_paths


def make_regionfiles(args, regions_dir):
    regionfile_path_list = list()
    if args.refver is not None:
        chromdict = common.ChromDict(refver=refver)
    elif args.fasta_path is not None:
        chromdict = common.ChromDict(fasta_path=fasta_path)

    merged_intvlist = workflow.get_merged_intervals_from_args(args)
    split_intvlists = merged_intvlist.get_split_interval_lists(
        chromdict, num=args.parallel)

    for zidx, intvlist in common.zenumerate(split_intvlists):
        outfile_path = os.path.join(regions_dir, f'{zidx}.bed')
        regionfile_path_list.append(outfile_path)
        intvlist.write_bed(outfile_path)

    return regionfile_path_list


def make_split_outfiles_subdirs(split_outfiles_dir, parallel):
    split_outfiles_subdirs = list()
    for zidx in common.zrange(parallel):
        subdir = os.path.join(split_outfiles_dir, zidx)
        os.mkdir(subdir)
        split_outfiles_subdirs.append(subdir)

    return split_outfiles_subdirs


def write_jobscripts(tmpdir_paths, split_outfiles_subdirs,

                        tbam_path
                        nbam_path
                        regionfile_path
                        split_outfile_subdir={repr(split_outfile_subdir)},
                        refver={repr(refver)},
                        fasta_path={repr(fasta_path)}, 
                        gcwiggle_path={repr(gcwiggle_path)}, 
                        hom={repr(hom)}, 
                        het={repr(het)},

                     jobname_prefix=PROGNAME, ncore_perjob=1):
    jobscript_path_list = list()

    for zidx, outfile_subdir in common.zenumerate(split_outfiles_subdirs):
        jobscript_path = os.path.join(tmpdir_paths['scripts'], 
                                      f'{zidx}.sbatch')
        jobscript_path_list.append(jobscript_path)
        logpath = os.path.join(tmpdir_paths['logs'], 
                               os.path.basename(jobscript_path) + '.log')
        success_logpath = re.sub('\.log$', '.success', logpath)
        failure_logpath = re.sub('\.log$', '.failure', logpath)

        script_contents = textwrap.dedent(f"""\
            #!{common.PYTHON}

            #SBATCH -N 1
            #SBATCH -n 1

            #SBATCH -c {ncore_perjob}
            #SBATCH -o {os.devnull}
            #SBATCH -e {os.devnull}
            #SBATCH -J {jobname_prefix}_{zidx}

            import os
            import contextlib
            import traceback
            import sys
            sys.path.append('{common.PACKAGE_LOCATION}')
            from {__name__} import unit_job

            log = open('{logpath}', 'w')
            with \\
                    contextlib.redirect_stdout(log), \\
                    contextlib.redirect_stderr(log):
                try:
                    unit_job(
                        tbam_path={repr(tbam_path)}, 
                        nbam_path={repr(nbam_path)}, 
                        regionfile_path={repr(regionfile_path)},
                        split_outfile_subdir={repr(split_outfile_subdir)},
                        refver={repr(refver)},
                        fasta_path={repr(fasta_path)}, 
                        gcwiggle_path={repr(gcwiggle_path)}, 
                        hom={repr(hom)}, 
                        het={repr(het)},
                        )
                except:
                    print(traceback.format_exc())
                    success = False
                else:
                    success = True
            log.close()

            if success:
                os.rename('{logpath}', '{success_logpath}')
            else:
                os.rename('{logpath}', '{failure_logpath}')
                raise SystemExit(1)
            """)

        with open(jobscript_path, 'w') as outfile:
            outfile.write(script_contents)

        os.chmod(jobscript_path, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

    return jobscript_path_list


def main(cmdargs):
    args = argument_parser(cmdargs)

    # setups logger
    logger = workflow.get_logger(name=PROGNAME, 
                                 stderr=(not args.silent),
                                 filenames=[args.log], append=False)
    logger.info('Beginning')

    # make 
    tmpdir_paths = make_tmpdir(args.tbam_path)
    regionfile_path_list = make_regionfiles(args, tmpdir_paths['regions_dir'])
    split_outfiles_subdirs = make_split_outfiles_subdirs(split_outfiles_dir, 
                                                         parallel)

    jobscript_path_list = write_jobscripts(
        tmpdir_paths, split_infile_path_list, fasta_path, 
        species, assembly, distance, refver,
        do_features, do_cosmic, do_popfreq,
        j)




import multiprocessing

import pandas as pd
import numpy as np
import pyranges as pr
import pysam

import handygenome.peakutils as peakutils
import handygenome.logutils as logutils
from handygenome.genomedf.genomedf import GenomeDataFrame
from handygenome.cnv.segment import SegmentDataFrame 
import handygenome.cnv.cncall as cncall
from handygenome.cnv.cncall import DefaultCNgUnavailableError


class DepthRawDataFrame(GenomeDataFrame):
    depth_colname = 'depth'
    norm_depth_colname = 'normalized_depth'
    MQ_colname = 'MQ'
    DEFAULT_DTYPES = {
        'Chromosome': 'string', 
        'Start': int, 
        'End': int,
        depth_colname: float,
        norm_depth_colname: float,
    }

    def init_sanitycheck(self):
        pass
        #assert self.__class__.depth_colname in self.annot_cols

    def spawn(self, frame):
        return super().spawn(frame, use_median=self.use_median)

    def __init__(self, refver, use_median=False):
        super().__init__(refver)
        self.use_median = use_median

    @classmethod
    def from_frame(cls, frame, refver, dtype=dict(), use_median=False):
        result = cls(refver, use_median=use_median)
        result.assign_frame(frame, dtype)
        result.init_sanitycheck()
        return result

    @classmethod
    def read_tsv(cls, filename, refver, use_median=False):
        """Column names must be set appropriately"""
        frame = cls.make_frame_from_tsv(filename)
        return cls.from_frame(frame, refver, use_median=use_median)

    @classmethod
    def load_mosdepth(cls, filename, refver, use_median=False):
        assert filename.endswith('.regions.bed.gz')
        frame = cls.make_frame_from_tsv(filename, annot_cols=[cls.depth_colname])
        return cls.from_frame(frame, refver, use_median=use_median)

    #################
    # normalization #
    #################

    def set_normalized_depth(self):
        all_depths = self.depth
        valid_idxs = ~np.isnan(all_depths)
        valid_depths = all_depths[valid_idxs]
        valid_lenghts = self.lengths[valid_idxs]
        average_depth = np.average(valid_depths, weights=valid_lenghts)

        self[self.__class__.norm_depth_colname] = all_depths / average_depth

    ################
    # segmentation #
    ################

    def get_segment(self, *args, **kwargs):
        seg = super().get_segment(*args, **kwargs)
        return DepthSegmentDataFrame.from_frame(seg.df, refver=self.refver)

    ##############
    # properties #
    ##############

    @property
    def depth(self):
        return self[self.__class__.depth_colname]

    @property
    def normalized_depth(self):
        return self[self.__class__.norm_depth_colname]

    norm_depth = normalized_depth

    @property
    def mq(self):
        return self[self.__class__.MQ_colname]

    ###############
    # MQ addition #
    ###############

    def get_MQ_fetchregion_gdf(self, readlength=151):
        self.sort()

        new_chroms = list()
        new_start0s = list()
        new_end0s = list()
        halflen = int(readlength / 2)

        for chrom, subgdf in self.group_bychrom(sort=False).items():
            start0s = subgdf.start0s
            end0s = subgdf.end0s
            lengths = subgdf.lengths

            long_length_selector = (lengths > readlength)
            start0s[long_length_selector] = start0s[long_length_selector] + halflen
            end0s[long_length_selector] = end0s[long_length_selector] - halflen

            new_start0s.append(start0s)
            new_end0s.append(end0s)
            new_chroms.append(np.repeat(chrom, len(start0s)))

        result_chroms = np.concatenate(new_chroms)
        result_start0s = np.concatenate(new_start0s)
        result_end0s = np.concatenate(new_end0s)

        return self.spawn_from_data(
            chroms=result_chroms, 
            start0s=result_start0s, 
            end0s=result_end0s,
        )

    def add_MQ(self, bam_path, readlength=151, nproc=1, n_split=1000, verbose=False):
        # get MQ values
        fetchregion_gdf = self.get_MQ_fetchregion_gdf(readlength)  # self is sorted in this step
        split_fetchregions = fetchregion_gdf.equal_nrow_split(n=n_split)
        args = (
            (bam_path, gdf.chroms, gdf.start0s, gdf.end0s, verbose)
            for gdf in split_fetchregions
        )
        with multiprocessing.Pool(nproc) as pool:
            map_result = pool.starmap(make_MQ_array, args)

        # apply results
        self[self.__class__.MQ_colname] = np.concatenate(map_result)

    def check_has_MQ(self):
        return self.__class__.MQ_colname in self.columns


def make_MQ_array(bam_path, chroms, start0s, end0s, verbose=False):
    if verbose:
        logutils.log(f'Beginning region chrom={chroms[0]}, start0={start0s[0]}, end0={end0s[0]}')
    with pysam.AlignmentFile(bam_path) as bam:
        def iterator():
            for c, s, e in zip(chroms, start0s, end0s):
                MQs = tuple(read.mapping_quality for read in bam.fetch(c, s, e))
                if len(MQs) == 0:
                    yield np.nan
                else:
                    yield np.mean(MQs)

        return np.fromiter(iterator(), dtype=float)


class DepthSegmentDataFrame(SegmentDataFrame):
    norm_depth_mean_colname = DepthRawDataFrame.norm_depth_colname + '_mean'
    norm_depth_std_colname = DepthRawDataFrame.norm_depth_colname + '_std'
    MQ_mean_colname = DepthRawDataFrame.MQ_colname + '_mean'

    @property
    def norm_depth_mean(self):
        return self[self.__class__.norm_depth_mean_colname]

    @property
    def norm_depth_std(self):
        return self[self.__class__.norm_depth_std_colname]

    def find_onecopy_depth(self, is_female, ploidy):
        try:
            default_CNg = self.get_default_CNg(is_female)
        except DefaultCNgUnavailableError:
            data = self.norm_depth_mean
            weights = self.lengths
        else:
            valid_selector = np.logical_and(
                (default_CNg != 0),
                ~np.isnan(default_CNg),
            )

            correction_factor = ploidy / default_CNg[valid_selector]
            data = self.norm_depth_mean[valid_selector] * correction_factor

            weights = self.lengths[valid_selector]

        # select copy-neutral depth segments
        peakresult = peakutils.find_density_peaks(
            data,
            weights=weights,
            xs=np.arange(0, 2, 0.01),
            bw_method=0.005,
        )
        modal_depth = peakresult['peak_xs'][np.argmax(peakresult['peak_ys'])]
        onecopy_depth = modal_depth / ploidy

        return onecopy_depth

    def add_rawdata_info(
        self, 
        depth_rawdata_gdf, 
        merge_methods=['mean', 'std'],
        rawdepth=False,
        nproc=1,
    ):
        depth_colname = (
            DepthRawDataFrame.depth_colname
            if rawdepth else
            DepthRawDataFrame.norm_depth_colname
        )
        right_gdf_cols = [depth_colname]

        MQ_colname = DepthRawDataFrame.MQ_colname
        if MQ_colname in depth_rawdata_gdf.columns:
            right_gdf_cols.append(MQ_colname)

        joined_gdf = self.drop_annots().join(
            depth_rawdata_gdf,
            right_gdf_cols=right_gdf_cols,
            how='left',
            merge=merge_methods,
            winsorize=(0.05, 0.05),
            nproc=nproc,
        )
        self.assign_frame(joined_gdf.df)

    @staticmethod
    def get_filter_colname():
        return f'depth_selected'

    def add_filter(self, MQ_cutoff=(50, None)):
        selector = self.filter_helper(self[self.__class__.MQ_mean_colname], MQ_cutoff)
        self[self.get_filter_colname()] = selector

    def get_filter(self):
        return self[self.get_filter_colname()]






import pyranges as pr
import numpy as np
import Bio.Seq
import Bio.SeqRecord

import handygenome.refgenome.refgenome as refgenome
import handygenome.read.readplus as libreadplus
import handygenome.align.msa as libmsa
import handygenome.align.alignhandler as alignhandler


def call_breakends(bam, chrom, start0, end0, refver, SA_region_factor=1):
    # make rpplist
    rpplist = libreadplus.ReadPlusPairList.from_bam(bam, chrom, start0, end0)

    # set params
    chromdict = refgenome.get_chromdict(refver)
    rp_dict = dict()
    clipspec_list = list()
    for rp, clipspec in rpplist.iter_clipspecs():
        rp_dict[rp.uid] = rp
        clipspec_list.append(clipspec)

    # get consensus sequences
    consensus_result = get_consensus_from_clipspecs(clipspec_list)

    # parse SA realignment regions
    SA_region_info = dict()
    for clipkey, sublist in consensus_result.items():
        # sanity check
        all_consensus_list = [x['consensus'] for x in sublist]
        assert len(all_consensus_list) == len(set(all_consensus_list))

        for consensus_item in sublist:
            rp_list = list()
            for seqrec in consensus_item['msa']:
                uid = seqrec.annotations['clipspec'].readuid
                rp = rp_dict[uid]
                rp_list.append(rp)

            confident_regions, unconfident_regions = get_SA_region_info(
                rp_list, chromdict, SA_region_factor,
            )

            key = clipkey + (consensus_item['true_consensus'],)
            SA_region_info[key] = {
                'confident': confident_regions,
                'unconfident': unconfident_regions,
                #'is_reversed': SAregion_is_reversed,
            }

    return SA_region_info, consensus_result


def get_SA_region_info(rp_list, chromdict, SA_region_factor):
    """Args:
        rp_list: list of ReadPlus objects which harbor a softclip of a given consensus
    """
    def make_gr_from_data(data):
        data_edit = tuple(zip(*data))
        gr = pr.from_dict(
            dict(Chromosome=data_edit[0], Start=data_edit[1], End=data_edit[2], is_reversed=data_edit[3])
        )

        # cluster
        gr = gr.cluster(slack=0)

        # merge by cluster; handle "is_reversed" values within a cluster
        merged_grs = list()
        for key, subdf in gr.df.groupby('Cluster'):
            flags = set(subdf['is_reversed'])
            if len(flags) == 1:
                is_reversed = flags.pop()
            else:
                is_reversed = None

            subgr = pr.PyRanges(subdf).merge()
            subgr.is_reversed = is_reversed
            merged_grs.append(subgr)

        # concat
        result = pr.concat(merged_grs)
        return result

    def extend_region(region, SA_region_factor, chromdict):
        """Modifies 'region' in-place"""
        length = region['end0'] - region['start0']
        pad = int(length * SA_region_factor)
        region['start0'] = max(0, region['start0'] - pad)
        region['end0'] = min(chromdict[region['chrom']], region['end0'] + pad)

    # collect SA region intervals
    confident_SA_regions = list()
    all_SA_regions = list()

    for rp in rp_list:
        n_clip = sum((x[0] == 4) for x in rp.read.cigartuples)
        is_confident = (n_clip == 1)  # This SAitem confidently matches the softclip

        for SAitem in rp.SAinfo:
            aligned_length = alignhandler.get_target_length(SAitem['cigartuples'])
            _start0 = SAitem['pos'] - 1
            _end0 = _start0 + aligned_length
            is_reversed = (SAitem['is_forward'] != rp.read.is_forward)
            SA_region = (SAitem['chrom'], _start0, _end0, is_reversed)

            all_SA_regions.append(SA_region)
            if is_confident:
                confident_SA_regions.append(SA_region)

    # merge SA regions, considering confident/nonconfident regions
    confident_regions = list()
    unconfident_regions = list()
    if all_SA_regions:
        # make into clustered grs
        all_gr = make_gr_from_data(all_SA_regions)
        if confident_SA_regions:
            confident_gr = make_gr_from_data(confident_SA_regions)
            for selector in np.eye(confident_gr.df.shape[0]).astype(bool):
                part_confident_gr = confident_gr[selector]
                joined = part_confident_gr.join(all_gr, report_overlap=True)
                if not joined.empty:
                    max_row = max(
                        (x[1] for x in joined.df.iterrows()),
                        key=(lambda x: x['Overlap']),
                    )
                    SA_region = {
                        'chrom': max_row['Chromosome'], 
                        'start0': max_row['Start_b'], 
                        'end0': max_row['End_b'], 
                        'is_reversed': max_row['is_reversed'],
                    }
                        # follows "is_reversed" value of the confident gr
                    confident_regions.append(SA_region)
        else:
            for idx, row in all_gr.df.iterrows():
                SA_region = {
                    'chrom': row['Chromosome'], 
                    'start0': row['Start'], 
                    'end0': row['End'], 
                    'is_reversed': row['is_reversed'],
                }
                unconfident_regions.append(SA_region)

    # extend merged SA regions
    for region in confident_regions:
        extend_region(region, SA_region_factor, chromdict)
    for region in unconfident_regions:
        extend_region(region, SA_region_factor, chromdict)

    return confident_regions, unconfident_regions



#############################################
# getting consensus from softclip fragments #
#############################################

def get_consensus_from_clipspecs(clipspec_list, aligner=alignhandler.ALIGNER_EQUAL_MM_GAP):
    # make SeqRecord objects from clipspecs
    seqrec_list = list()
    for clipspec in clipspec_list:
        seqrec_seq = (clipspec.seq if clipspec.is_forward else clipspec.seq[::-1])
        seqrec_id = f'(qname={clipspec.readuid.qname}, pos0={clipspec.pos0:,}, is_forward={clipspec.is_forward})'
        seqrec = Bio.SeqRecord.SeqRecord(
            seq=Bio.Seq.Seq(seqrec_seq), 
            id=seqrec_id, 
            annotations={'clipspec': clipspec},
        )
        seqrec_list.append(seqrec)

    # group seqrecs by clipspec identity
    clipid_groups = dict()
    for seqrec in seqrec_list:
        clipspec = seqrec.annotations['clipspec']
        clipid = (clipspec.readuid.chrom, clipspec.pos0, clipspec.is_forward)
        clipid_groups.setdefault(clipid, list())
        clipid_groups[clipid].append(seqrec)

    # make consensus seqs for each clipspec group
    result = dict()
    for clipid, sublist in clipid_groups.items():
        result[clipid] = list()
        seqrec_groups = group_seqrecs(sublist, aligner, score_limit=0.7)
        for target, subseq_info in seqrec_groups.items():
            target_seqrec = subseq_info['superseq']['seqrec']
            query_seqrecs = [x['seqrec'] for x in subseq_info['subseqs']]
            alignments = [x['alignment'] for x in subseq_info['subseqs']]

            # get consensus
            msa, consensus = libmsa.msa_with_target(
                target, alignments, 
                target_seqrec=target_seqrec, 
                query_seqrecs=query_seqrecs
            )
            true_consensus = (consensus if clipid[2] else consensus[::-1])

            # result
            result[clipid].append(
                {
                    'target': target, 
                    'msa': msa, 
                    'consensus': consensus, 
                    'true_consensus': true_consensus,
                }
            )

    return result


def group_seqrecs(seqrec_list, aligner, score_limit=0.7):
    """Groups SeqRecord objects according to sequence similarity"""
    def init_group(superseq_seqrec):
        return {
            'superseq': {
                'seqrec': superseq_seqrec, 'alignment': None, 'adj_score': None,
            },
            'subseqs': list(),
        }

    groups = dict()
    for seqrec in sorted(seqrec_list, key=len, reverse=True):
        seq = str(seqrec.seq)
        if len(groups) == 0:
            groups[seq] = init_group(seqrec)
            #groups[seqrec.seq].append({'seqrec': seqrec, 'alignment': None, 'adj_score': None})
            continue

        # find superseq candidates
        superseq_matching_info = list()
        for superseq in groups.keys():
            aln = alignhandler.alignment_tiebreaker(aligner.align(superseq, seqrec.seq))
            adj_score = score_alignment(aln)
            superseq_matching_info.append((superseq, aln, adj_score))

        # pick the most similar superseq
        chosen_info = max(superseq_matching_info, key=(lambda x: x[2]))
        superseq, aln, adj_score = chosen_info

        if adj_score >= score_limit:
            groups[superseq]['subseqs'].append(
                {'seqrec': seqrec, 'alignment': aln, 'adj_score': adj_score}
            )
        else:
            assert seq not in groups
            groups[seq] = init_group(seqrec)

    return groups


def score_alignment(alignment, match_score=1, mismatch_score=0, gap_score=0):
    # get walks
    walks = alignhandler.get_walks(alignment, copy=True)
    if not walks[0].check_both():
        del walks[0]
    if not walks[-1].check_both():
        del walks[-1]

    # calc score
    #target_seq = np.array(tuple(alignment.target))
    #query_seq = np.array(tuple(alignment.query))
    target_seq = np.fromiter(iter(alignment.target), dtype='<U1')
    query_seq = np.fromiter(iter(alignment.query), dtype='<U1')
    score = 0
    for walk in walks:
        if walk.check_both():
            target_subseq = target_seq[walk.get_target_slice()]
            query_subseq = query_seq[walk.get_query_slice()]

            n_match = (target_subseq == query_subseq).sum()
            n_mismatch = len(walk) - n_match

            score += n_match * match_score
            score += n_mismatch * mismatch_score
        else:
            score += len(walk) * gap_score

    # make adjusted score
    aligned_length = sum(len(x) for x in walks)
    adj_score = score / aligned_length

    return adj_score





import textwrap

import pysam

import importlib
top_package_name = __name__.split('.')[0]
common = importlib.import_module('.'.join([top_package_name, 'common']))
infoformat = importlib.import_module('.'.join([top_package_name, 'variant', 'infoformat']))


class SomaticIndex:
    meta = {'ID': 'SomaticIndex', 'Type': 'Integer', 'Number': 1,
            'Description': (
                'The index of the allele corresponding to somatic mutation. '
                '0 means REF, 1 means the first ALT allele, 2 means the '
                'second ALT allele, and so on. It is 1 in most cases.')}
    type = 'INFO'

    def __init__(self, index=1):
        self.index = index

    def __repr__(self):
        return f'<SomaticIndex={self.index}>'

    @classmethod
    def from_vr(cls, vr):
        val = infoformat.get_info(vr, cls.meta['ID'], collapse_tuple=False)
        return cls(val)

    @classmethod
    def write_meta(cls, vcfheader):
        vcfheader.add_meta(key=cls.type, items=cls.meta.items())
        
    def write(self, vr):
        infoformat.set_info(vr, self.__class__.meta['ID'], self.index,
                            typeconv=False)
        

class GermlineIndexes:
    meta = {'ID': 'GermlineIndexes', 'Type': 'Integer', 'Number': '.',
            'Description': (
                'The indexes of the germline alleles. '
                '0 means REF, 1 means the first ALT allele, 2 means the '
                'second ALT allele, and so on. It is 0,0 in most cases.')}
    type = 'INFO'

    def __init__(self, indexes=(0, 0)):
        self.indexes = tuple(indexes)
        
    def __repr__(self):
        return f'<GermlineIndexes={self.indexes}>'

    @classmethod
    def from_vr(cls, vr):
        val = infoformat.get_info(vr, cls.meta['ID'], collapse_tuple=False)
        return cls(val)
    
    @classmethod
    def write_meta(cls, vcfheader):
        vcfheader.add_meta(key=cls.type, items=cls.meta.items())

    def write(self, vr):
        infoformat.set_info(vr, self.__class__.meta['ID'], self.indexes,
                            typeconv=False)

    def write_GT(self, vr, sampleid):
        vr.samples[sampleid].allele_indices = self.indexes
        vr.samples[sampleid].phased = False



import pprint
import os
import re
import gzip
import shutil
import itertools
import tempfile
import collections
import inspect
import functools
import math

import numpy as np
import pandas as pd
import scipy.stats
import scipy.signal
import scipy.linalg
import Bio.bgzf
import matplotlib as mpl
import matplotlib.pyplot as plt

import handygenome.deco as deco
import handygenome.logutils as logutils


PAT_INT = re.compile('-?[0-9]+')
PAT_FLOAT = re.compile('(-?[0-9]+\.[0-9]+)|(-?[0-9]+(\.[0-9]+)?e-?[0-9]+)')


#####################
# custom class base #
#####################

def repr_base(obj, keylist, comma_sep_int=False):
    string_list = list()
    for key in keylist:
        val = getattr(obj, key)
        if isinstance(val, int) and comma_sep_int:
            string_list.append(f'{key}={val:,}')
        else:
            string_list.append(f'{key}={repr(val)}')

    return ', '.join(string_list)


##########
# colors #
##########

COLORS = {
    'red':     '\033[38;5;196m',
    'magenta': '\033[38;5;201m',
    'pink':    '\033[38;5;213m',
    'orange':  '\033[38;5;9m',
    'yellow':  '\033[38;5;214m',
    'gold':    '\033[38;5;11m',
    'green':   '\033[38;5;40m',
    'blue':    '\033[38;5;33m',
    'cyan':    '\033[38;5;14m',
    'purple':  '\033[38;5;93m',
    'gray':    '\033[38;5;8m',
    'white':   '\033[38;5;15m',
    'end':     '\033[0m',
}


class ColorsQQ:
    mine="\033[48;5;6m"
    busy="\033[48;5;244m"
    free="\033[48;5;238m"
    end="\033[0m"
    nor="\033[48;5;160m"
    nor2="\033[48;5;52m"
    danger1="\033[38;5;208m"
    danger2="\033[38;5;196m"
    darkgray="\033[38;5;240m"


def visualize_colors():
    for i in range(256):
        print(f'{i:<3d} \033[38;5;{i}m\\033[38;5;{i}m\033[0m')


def cpformat(obj, **kwargs):
    result = pprint.pformat(obj, **kwargs)
    result = re.sub('(True)', COLORS['green'] + '\\1' + COLORS['end'], result)
    result = re.sub('(False)', COLORS['red'] + '\\1' + COLORS['end'], result)
    result = re.sub('(None)', COLORS['purple'] + '\\1' + COLORS['end'], result)
    return result


def cpprint(obj):
    print(cpformat(obj))


#######################
# string manipulators #
#######################

def str_to_nonstr(val):
    #assert isinstance(val, str)

    if val.lower() in ('none', 'null'):
        return None
    elif val.lower() == 'nan':
        return np.nan
    elif val.lower() == 'true':
        return True
    elif val.lower() == 'false':
        return False
    elif PAT_INT.fullmatch(val) is not None:
        return int(val)
    elif PAT_FLOAT.fullmatch(val) is not None:
        return float(val)
    else:
        return val


def nonstr_to_str(val):
    assert not isinstance(val, str)

    if val is None:
        return None
    else:
        return str(val)


def zrange(n):
    assert n > 0
    width = len(str(n - 1))
    for idx in range(n):
        zidx = str(idx).zfill(width)
        yield zidx


def zenumerate(iterable):
    iterable_tup = tuple(iterable)
    length = len(iterable_tup)
    for zidx, item in zip(zrange(length), iterable_tup):
        yield zidx, item


def zidx_to_idx(zidx):
    if re.match('^0+$', zidx):
        return 0
    else:
        return int(re.sub(r'^0*', '', zidx))


def get_padded_indices(n):
    """Begins with 0"""
    width = len(str(n-1))
    for idx in range(n):
        yield str(idx).zfill(width)


def shorten_int(numlist, n_after_dot=3):
    mapping = {0: '', 1: 'K', 2: 'M', 3: 'G'}

    log = np.log10(numlist)
    assert (log < 12).all(), f'Numbers greater than or equal to 10^12 are not allowed.'

    #qs = [int(y) for y in (log / 3)]
    qs = np.floor(log / 3)

    suffixes = [mapping[x] for x in qs]
    #new_numlist = numlist / (10 ** (np.array(qs) * 3))
    new_numlist = numlist / (10 ** (qs * 3))
    formatter = f'.{n_after_dot}f'
    result = [
        (
            f'{int(x)}' 
            if y == '' else 
            f'{{x:{formatter}}} {{y}}'.format(x=x, y=y)
        )
        for x, y in zip(new_numlist, suffixes)
    ]
    return result


'''
http://code.activestate.com/recipes/578019
https://psutil.readthedocs.io/en/latest/#bytes-conversion
'''
_bytes2human_symbols = ('K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
_bytes2human_symbols_rev = tuple(reversed(_bytes2human_symbols))
_bytes2human_prefix = {
    s: 1 << (i + 1) * 10
    for i, s in enumerate(_bytes2human_symbols)
}

def bytes2human(n):
    for s in _bytes2human_symbols_rev:
        if abs(n) >= _bytes2human_prefix[s]:
            value = float(n) / _bytes2human_prefix[s]
            return '%.1f%s' % (value, s)
    return "%sB" % n


# listdir

def listdir(path):
    return sorted(os.path.join(path, x) for x in os.listdir(path))


#################
# file handling #
#################

def unzip(src, dest, rm_src=False):
    with gzip.open(src, 'rb') as infile:
        with open(dest, 'wb') as outfile:
            shutil.copyfileobj(infile, outfile)      
    if rm_src:
        os.remove(src)


#def check_file_is_plain(fname):
def check_textfile(fname):
    with open(fname, 'r') as f:
        try:
            _ = f.read(1)
        except UnicodeDecodeError:
            return False
        else:
            return True


def check_gzipped(fname):
    with gzip.open(fname, 'rb') as f:
        try:
            _ = f.read(1)
        except gzip.BadGzipFile as exc:
            if str(exc).startswith('Not a gzipped file'):
                return False
            else:
                raise
        else:
            return True


def check_bgzipped(fname):
    try:
        f = Bio.bgzf.open(fname)
    except ValueError as exc:
        if str(exc).startswith('A BGZF (e.g. a BAM file) block should start with'):
            return False
        else:
            raise
    else:
        f.close()
        return True


def openfile(fname, mode='rt'):
    patstring = r'[rwa][tb]'
    if not re.fullmatch(patstring, mode):
        raise Exception(f'Invalid "mode" argument. Allowed pattern: {patstring}')

    if len(mode) == 1:
        mode = mode + 't'

    if check_textfile(fname):
        return open(fname, mode)
    elif check_gzipped(fname):
        return gzip.open(fname, mode)
    else:
        raise Exception(f'Input file is neither plain text nor gzipped.')


def fileiter(path, sep='\t', remove_leading_hashes=True, skip_double_hashes=True):
    infile = openfile(path, 'r')

    while True:
        line = next(infile)
        if skip_double_hashes:
            if line.startswith('##'):
                continue
            else:
                break
        else:
            break
        
    headerline = line
    if remove_leading_hashes:
        headerline = re.sub('^#*', '', headerline)
    header = get_linesp(headerline, sep=sep)

    for line in infile:
        linesp = get_linesp(line, sep=sep)
        if len(linesp) != len(header):
            raise Exception(
                f'Field numbers of the header line and the current '
                f'line are different:\n'
                f'header: {header}\n'
                f'line: {linesp}')
        linedict = dict(zip(header, linesp))
        yield linedict

    infile.close()


def rm_newline(line):
    return re.sub('(\r)?\n$', '', line)


def rm_newline_byte(byteline):
    return re.sub(b'(\r)?\n$', b'', byteline)


def get_linesp(line, sep='\t'):
    return rm_newline(line).split(sep)


def get_linesp_byte(byteline, sep=b'\t'):
    return rm_newline_byte(byteline).split(sep)


# printwidth

def printwidth_get_width_list(df):
    '''
    df: [
    [line1_field1, line1_field2, ... ],
    [line2_field1, line2_field2, ... ],
    ...,
    ]
    '''
    width_list = list()
    for i in range(len(df[0])):
        width_list.append(list())

    for line in df:
        for idx, field in enumerate(line):
            width_list[idx].append(len(field))

    for idx, e in enumerate(width_list):
        width_list[idx] = max(e)

    return width_list


def printwidth_print_line(line, width_list, margin, target):
    printresult = ''
    for idx, e in enumerate(line):
        printresult += f'{e:>{width_list[idx] + margin}s}'
    if target == 'out':
        print(printresult, flush = True)
    elif target == 'err':
        print(printresult, flush = True, file = sys.stderr)


def printwidth(df, margin = 2, target = 'out'):
    for line in df:
        for idx, e in enumerate(line):
            line[idx] = str(e)

    width_list = printwidth_get_width_list(df)
    for line in df:
        printwidth_print_line(line, width_list, margin, target)


######################
# memory usage check #
######################

def get_rss(mode='total', unit='b'):
    assert mode in ('self', 'children', 'total')
    assert unit in ('b', 'k', 'm', 'g')

    proc = psutil.Process()
    children = proc.children(recursive=True)
    rss_self = proc.memory_info().rss  # in bytes
    rss_children = sum(p.memory_info().rss for p in children)  # in bytes

    if mode == 'self':
        rss = rss_self
    elif mode == 'children':
        rss = rss_children
    elif mode == 'total':
        rss = rss_self + rss_children

    if unit == 'b':
        rss = rss
    elif unit == 'k':
        rss = rss / 1024
    elif unit == 'm':
        rss = rss / 1024**2
    elif unit == 'g':
        rss = rss / 1024**3

    return rss


#######################
# iteration utilities #
#######################

def _multi_selector_base(sequence, comparing_func, key=None, with_target_val=False):
    sequence = list(sequence)
    if key is None:
        values = sequence
    else:
        values = [key(x) for x in sequence]
    target_val = comparing_func(values)
    is_target_val = [x == target_val for x in values]

    if with_target_val:
        return list(itertools.compress(sequence, is_target_val)), target_val
    else:
        return list(itertools.compress(sequence, is_target_val))


def multi_min(sequence, key=None, with_target_val=False):
    return _multi_selector_base(sequence, min, key=key, with_target_val=with_target_val)


def multi_max(sequence, key=None, with_target_val=False):
    return _multi_selector_base(sequence, max, key=key, with_target_val=with_target_val)


# https://stackoverflow.com/questions/480214/how-do-i-remove-duplicates-from-a-list-while-preserving-order
def unique_keeporder(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


#########################################
# utilities from itertools plus package #
#########################################

def pairwise(iterable):
    iterable = iter(iterable)
    try:
        x0 = next(iterable)
        x1 = next(iterable)
    except StopIteration:
        return

    while True:
        yield (x0, x1)
        try:
            x2 = next(iterable)
        except StopIteration:
            return
        x0 = x1
        x1 = x2
        continue


def chunk_iter(iterable, n):
    args = [iter(iterable)] * n
    for subiter in itertools.zip_longest(*args, fillvalue=None):
        yield tuple(x for x in subiter if x is not None)


# from Itertools Recipes; similar as chunk_iter
def grouper_Itertools_Recipes(iterable, n, *, incomplete='fill', fillvalue=None):
    "Collect data into non-overlapping fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, fillvalue='x') --> ABC DEF Gxx
    # grouper('ABCDEFG', 3, incomplete='strict') --> ABC DEF ValueError
    # grouper('ABCDEFG', 3, incomplete='ignore') --> ABC DEF
    args = [iter(iterable)] * n
    if incomplete == 'fill':
        return itertools.zip_longest(*args, fillvalue=fillvalue)
    if incomplete == 'strict':
        return zip(*args, strict=True)
    if incomplete == 'ignore':
        return zip(*args)
    else:
        raise ValueError('Expected fill, strict, or ignore')



######################################
# array handlers & mathematical ones #
######################################

# Proposed by: Noyer282 (https://stackoverflow.com/a/40426159/7204581)
# https://stackoverflow.com/questions/432112/is-there-a-numpy-function-to-return-the-first-index-of-something-in-an-array
def array_index(arr, v):
    assert arr.ndim == 1
    return next((idx for idx, val in enumerate(arr) if val == v), None)


def round_sig(num, n):
    """round 'num' with 'n' significant digits"""

    assert n > 0
    return round(num, -math.floor(math.log10(abs(num))) + (n-1))


def get_split_nums_byN(total_length, N):
    arrsp = np.array_split(np.zeros(total_length), N)
    return np.fromiter(
        (x.shape[0] for x in arrsp if x.shape[0] != 0), 
        dtype=int,
    )


def get_split_nums_byN_type2(total_length, N):
    total_num = min(N, total_length)
    q, r = divmod(total_length, total_num)
    width1 = q + 1
    num1 = r
    width2 = q
    num2 = total_num - r

    return np.repeat([width1, width2], [num1, num2])


def get_split_nums_bywidth(total_length, width):
    width_raw = min(width, total_length)
    q, r = divmod(total_length, width_raw)
    if r == 0:
        width = width_raw
        num = q
        result = np.repeat(width, num)
    else:
        q2, r2 = divmod(r, q)
        width1 = width_raw + q2 + 1
        num1 = r2
        width2 = width_raw + q2
        num2 = q - r2
        result = np.repeat([width1, width2], [num1, num2])

    return result


def array_grouper(arr, omit_values=False):
    """Does not sort before grouping, like itertools.groupby"""
    assert arr.ndim in (1, 2)

    diff = np.empty(arr.shape[0], dtype=bool)
    diff[0] = True
    if arr.ndim == 1:
        diff[1:] = np.diff(arr)
    elif arr.ndim == 2:
        diff[1:] = np.diff(arr, axis=0).any(axis=1)

    indexes = np.nonzero(diff)[0]

    if omit_values:
        values = None
    else:
        values = arr[indexes]

    counts = np.empty(indexes.shape, dtype=int)
    counts[:-1] = np.diff(indexes)
    counts[-1] = arr.shape[0] - indexes[-1]

    groupkey = np.repeat(np.arange(len(counts)), counts)
        
    return values, counts, groupkey
                

def get_indexes_of_array(values, ordered_keys):
    """Used for handling pandas.cut function output, in cnv.misc module"""
    result = np.zeros(len(values))
    for idx, key in enumerate(ordered_keys):
        result[np.where(values == key)[0]] = idx

    return result


def nanaverage(values, weights):
    #assert isinstance(values, np.ndarray)
    #assert isinstance(weights, np.ndarray)

    values = np.array(values)
    weights = np.array(weights)

    selector = ~np.isnan(values)
    new_values = values[selector]
    new_weights = weights[selector]
    return np.average(new_values, weights=new_weights)


def get_ranks(arr):
    ranks = scipy.stats.rankdata(data, method='max')
    return ranks / len(ranks)


def get_mode(data, xs=np.arange(0, 0.51, 0.01)):
    kernel = scipy.stats.gaussian_kde(data)
    ys = kernel(xs)
    return xs[np.argmax(ys)]


def bernoulli_iterator(iterable, p, block_size=int(1e5)):
    """Args:
        p: probability of selection
    """
    for x, rv in zip(iter(iterable), bernoulli_rv_generator(p, block_size=block_size)):
        if rv:  # True if 1
            yield x


def bernoulli_rv_generator(p, block_size=int(1e5)):
    while True:
        rvs = scipy.stats.bernoulli.rvs(p=p, loc=0, size=block_size)
        for x in rvs:
            yield x


def mean_mad(values):
    values = np.array(values)
    return np.mean(np.abs(values - np.mean(values)))


def median_mad(values):
    values = np.array(values)
    return np.median(np.abs(values - np.median(values)))


def get_diffmean(values, weights=None):
    assert values.ndim == 1

    if weights is None:
        weights = np.ones_like(values)
    else:
        assert weights.shape == values.shape

    indexes = np.triu_indices(values.shape[0], k=1)
    diffs = values[indexes[0]] - values[indexes[1]]
    diff_weights = weights[indexes[0]] + weights[indexes[1]]

    return np.average(np.abs(diffs), weights=diff_weights)


def dirichlet_multinomial_rvs(n, alpha, rng=None):
    """Args:
        n: int or 1d array. Number of trials.
        alpha: alpha parameter passed to Dirichlet distribution.

    A single Dirichlet distribution is created with "alpha" parameter.
    Proportions for multinomial distribution are drawn from the Dirichlet 
    distribution for each element of "n" parameter.

    Returns:
        array with shape (len(n), len(alpha))
    """
    n = np.atleast_1d(n)
    assert n.ndim == 1

    if rng is None:
        rng = np.random.default_rng()

    props = rng.dirichlet(alpha=alpha, size=n.shape[0])
    result = rng.multinomial(n=n, pvals=props)
    return result


def get_nearest_integer_bounds(x, elevate=True):
    x = np.atleast_1d(x)

    lower = np.floor(x).astype(int)
    upper = np.ceil(x).astype(int)
    same_indexes = (upper == lower)
    if elevate:
        upper[same_indexes] += 1
    else:
        lower[same_indexes] -= 1

    return upper, lower


##############################
# Genomic coordinate sorting #
##############################

def coord_sortkey(chrom, pos, chromdict): 
    """
    Args:
        pos: 1-based
        chromdict: ChromDict class instance
    """

    return (chromdict.contigs.index(chrom), pos)


#def get_vcfspec_sortkey(chromdict):
#    def sortkey(vcfspec):
#        return (chromdict.contigs.index(vcfspec.chrom), vcfspec.pos, vcfspec.ref) + vcfspec.alts
#        #return coord_sortkey(vcfspec.chrom, vcfspec.pos, chromdict)
#
#    return sortkey


def compare_coords(chrom1, pos1, chrom2, pos2, chromdict):
    """
    Returns:
        0: equal; -1: chrom1/pos1 comes first; 1: chrom2/pos2 comes first
    """

    if chrom1 == chrom2 and pos1 == pos2:
        return 0
    else:
        if (coord_sortkey(chrom1, pos1, chromdict) 
            < coord_sortkey(chrom2, pos2, chromdict)):
            return -1
        else:
            return 1


##################################
# temporary file path generation #
##################################

def get_tmpfile_path(prefix=None, suffix=None, dir=None, where=None, delete=False, is_dir=False):
    """Args:
        where: alias for dir
    """
    # alias handling
    assert not ((dir is not None) and (where is not None))
    if where is not None:
        dir = where
    # main
    if dir is None:
        dir = os.getcwd()

    if delete:
        fd, path = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=dir)
        os.close(fd)
        os.remove(path)
    else:
        if is_dir:
            path = tempfile.mkdtemp(prefix=prefix, suffix=suffix, dir=dir)
        else:
            fd, path = tempfile.mkstemp(prefix=prefix, suffix=suffix, dir=dir)
            os.close(fd)

    return path


#####################
# callable handling #
#####################

def check_is_lambda(obj):
    return (
        callable(obj)
        and (getattr(obj, '__name__') == '<lambda>')
    )


##################
# set operations #
##################

def check_unique(seq):
    seq = tuple(seq) 
    seq_set = set(seq)
    return len(seq) == len(seq_set)





import pandas as pd
import numpy as np

import rpy2.robjects as ro
from rpy2.robjects.packages import importr
from rpy2.robjects import pandas2ri


'''rpy2 package seems to utilize 'root' logger. 
In logutils module, when 'root' logger is used, ("log_old" function), rpy2 emits some log messages.
When logutils module use a logger with another name, (current "log" function), the rpy2 log message is not seen anymore.
'''


DNACOPY = importr('DNAcopy')


#def run_segmentation(
#    gdf, 
#    colname='value', 
#    out_colname=None, 
#    only_coords=True,
#    #join_raw=False,
#    #N_colname='N',
#
#    smoothing=False, 
#    verbose=True, 
#    smooth_kwargs=dict(), 
#    segment_kwargs=dict(),
#):
#    """kwargs for 'DNAcopy::segment' function (https://bioconductor.org/packages/release/bioc/manuals/DNAcopy/man/DNAcopy.pdf)
#        - weights
#        - alpha
#        - nperm
#        - p.method
#        - min.width
#        - kmax
#        - nmin
#        - eta
#        - sbdry
#        - trim
#        - undo.splits
#        - undo.prune
#        - undo.SD
#        - verbose
#        
#    - Input position to R dnacopy is calculated as floor((start0 + end0 - 1) / 2)
#    """ 
#    # sanitycheck & argument handling
#    assert colname in gdf.columns
#    if out_colname is None:
#        out_colname = colname
#
#    # run R function
#    seg_df = run_dnacopy_segment(
#        chromosomes=gdf['Chromosome'], 
#        positions=gdf.get_midpoints(), 
#        values=gdf[colname],
#
#        smoothing=smoothing, 
#        verbose=verbose, 
#        smooth_kwargs=smooth_kwargs,
#        segment_kwargs=segment_kwargs,
#    )
#
#    # postprocess df
#    seg_df.drop('ID', axis=1, inplace=True)
#    seg_df.rename(
#        columns={
#            'chrom': 'Chromosome', 
#            'loc.start': 'Start',  # 0-based closed interval
#            'loc.end': 'End',  # 0-based closed interval
#            'num.mark': 'num_data', 
#            'seg.mean': out_colname,
#        }, 
#        inplace=True
#    )
#    seg_df['End'] = seg_df['End'] + 1  # to make into 0-based half-open system
#
#    if only_coords:
#        seg_df.drop(['num_data', out_colname], axis=1, inplace=True)
#
#    return seg_df
#
#    #result = GDF.from_frame(seg_df, gdf.refver)
#
##    if join_raw:
##        result = result.join(
##            gdf, 
##            other_cols=[colname], 
##            how='left', 
##            merge=['mean', 'std'], 
##            ddof=0,
##            overlapping_length=False,
##            N_colname=N_colname,
##        )
#
#    #return result


def rdataframe_to_df(rdataframe):
    with (ro.default_converter + pandas2ri.converter).context():
        return ro.conversion.get_conversion().rpy2py(rdataframe)


def run_dnacopy_segment(
    chromosomes, positions, values,
    *,
    N_colname='N',
    value_colname='value',
    smoothing=False, 
    verbose=True, 
    smooth_kwargs=dict(), 
    segment_kwargs=dict(),
):
    """kwargs for 'DNAcopy::segment' function (https://bioconductor.org/packages/release/bioc/manuals/DNAcopy/man/DNAcopy.pdf)
        - weights
        - alpha
        - nperm
        - p.method
        - min.width
        - kmax
        - nmin
        - eta
        - sbdry
        - trim
        - undo.splits
        - undo.prune
        - undo.SD
        - verbose
        
    - Works well when 'positions' argument contains 0 or negative integers.
    """ 
    # prepare R function arguments
    default_segment_kwargs = {
        'alpha': 0.01,
        'nperm': 10000,
        'p_method': 'hybrid',
        'min_width': 2,
        'kmax': 25,
        'nmin': 200,
        'eta': 0.05,
        'trim': 0.025,
        'undo_splits': 'none',
        'undo_prune': 0.05,
        'undo_SD': 3,
    }
    assert set(segment_kwargs.keys()).issubset(default_segment_kwargs.keys())
    segment_kwargs = default_segment_kwargs | segment_kwargs
    segment_kwargs['verbose'] = int(verbose)

    default_smooth_kwargs = {
        'smooth_region': 10, 
        'outlier_SD_scale': 4, 
        'smooth_SD_scale': 2, 
        'trim': 0.025, 
    }
    assert set(smooth_kwargs.keys()).issubset(default_smooth_kwargs.keys())
    smooth_kwargs = default_smooth_kwargs | smooth_kwargs

    # main
    arg_chrom = ro.StrVector(chromosomes)
    arg_pos = ro.IntVector(positions)
    arg_value = ro.FloatVector(values)

    cnaobj = DNACOPY.CNA(arg_value, arg_chrom, arg_pos, data_type='logratio')
    if smoothing:
        cnaobj = DNACOPY.smooth_CNA(cnaobj, **smooth_kwargs)
    segresult = DNACOPY.segment(cnaobj, **segment_kwargs)

    # turn into pandas dataframe
    segdf = rdataframe_to_df(segresult.rx2['output'])
    # segdf is an R dataframe which has following columns: 
        # ID, chrom, loc.start, loc.end, num.mark, seg.mean

    segdf.drop('ID', axis=1, inplace=True)
    segdf.rename(
        columns={
            'chrom': 'Chromosome', 
            'loc.start': 'Start',  # 0-based closed interval
            'loc.end': 'End',  # 0-based closed interval
            'num.mark': N_colname, 
            'seg.mean': value_colname,
        }, 
        inplace=True
    )
    segdf['End'] = segdf['End'] + 1  # to make into 0-based half-open system

    return segdf

    #return {
    #    'Chromosome': np.array(segdf.rx2['chrom']),
    #    'Start': np.array(segdf.rx2['loc.start']),
    #    'End': (np.array(segdf.rx2['loc.end']) + 1),
    #    'N': np.array(segdf.rx2['num.mark']),
    #    'value': np.array(segdf.rx2['seg.mean']),
    #}





import os
import warnings

import pysam

import handygenome.variant.varianthandler as varianthandler


DEFAULT_VCFVER = '4.3'


def write_empty_vcf(outfile_path, chromdict=None, samples=None, pysamhdr=None, 
                    vcfver=DEFAULT_VCFVER):
    """
    Args:
        outfile_path: Output vcf file path.
        chromdict: Designed as a mandatory one since vcf header without 
            contig metadata may result in segmentation fault with writing 
            to a file.
        samples: Iterable containing sample names.
        pysamhdr: pysam.VariantHeader object. All metadata from this, 
            except fileformat and contig, is used.
        vcfver: Default 4.3
    """

    with open(outfile_path, 'w') as outfile:
        outfile.write(f'##fileformat=VCFv{vcfver}\n') # fileformat

        if chromdict is not None:
            for chrom, length in chromdict.items(): # contig
                outfile.write(f'##contig=<ID={chrom},length={length}>\n')

        if pysamhdr is not None: # other metadata
            for hdr_rec in pysamhdr.records:
                if chromdict is not None:
                    if (
                            (hdr_rec.type != 'CONTIG') and 
                            (hdr_rec.key != 'fileformat')):
                        outfile.write(str(hdr_rec))
                else:
                    if hdr_rec.key != 'fileformat':
                        outfile.write(str(hdr_rec))

        headerline = ['#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 
                      'INFO']
        if samples is not None:
            headerline.append('FORMAT')
            for sample in samples:
                headerline.append(sample)
        outfile.write('\t'.join(headerline) + '\n')


def create_empty_pysamvcf(chromdict=None, samples=None, pysamhdr=None, 
                          vcfver=DEFAULT_VCFVER):
    """
    Args:
        outfile_path: Output vcf file path.
        chromdict: Designed as a mandatory one since vcf header without 
            contig metadata may result in segmentation fault with writing to 
            a file.
        samples: Iterable containing sample names.
        pysamhdr: pysam.VariantHeader object. All metadata from this, except 
            fileformat and contig, is used.
        vcfver: Default 4.3
    """

    tf_path = workflow.get_tmpfile_path(delete=False)
    write_empty_vcf(tf_path, chromdict, samples, pysamhdr, vcfver)
    empty_vcf = pysam.VariantFile(tf_path, mode = 'r')
    os.remove(tf_path)

    return empty_vcf


######################################

def create_header(chromdict, samples=None, vcfheader=None):
    """
    Args:
        chromdict: Designed as a mandatory one since vcf header without 
            contig metadata may result in segmentation fault with writing 
            to a file.
        samples: An iterable containing sample names.
        vcfheader: Another pysam.VariantHeader object. All metadata from 
            this, except fileformat and contig, is used.

    Returns:
        A pysam.VariantHeader object
    """

    result = pysam.VariantHeader()

    for chrom, length in chromdict.items():
        result.contigs.add(chrom, length)

    if samples is not None:
        for sampleid in samples:
            result.samples.add(sampleid)

    if vcfheader is not None:
        for rec in vcfheader.records:
            if (rec.type != 'CONTIG') and (rec.key != 'fileformat'):
                result.add_record(rec)

    return result


#def get_updated_header(vcfhdr, chromdict = None, samples = None):
#    new_header = create_header(chromdict, samples, vcfhdr)
#    for header_rec in vcfhdr.records:
#        if ( header_rec.type != 'CONTIG' ) and ( header_rec.key != 'fileformat' ):
#            new_header.add_record(header_rec)
#
#    return new_header


def create_vr(chromdict, *, vcfspec=None, end=None, samples=None, vcfheader=None, use_header=False):
    if use_header:
        vr = vcfheader.new_record()
    else:
        hdr = create_header(chromdict, samples, vcfheader)
        vr = hdr.new_record()

    if vcfspec is not None:
        vcfspec.apply_to_vr(vr)
        #varianthandler.apply_vcfspec(vr, vcfspec)

    if end is not None:
        vr.stop = end

    return vr



import re
import os
import datetime
import inspect
import itertools
import contextlib
import multiprocessing
import functools

import pandas as pd
import numpy as np
import pyranges as pr
import scipy.stats

import handygenome.deco as deco
import handygenome.refgenome.refgenome as refgenome
import handygenome.tools as tools
import handygenome.logutils as logutils
import handygenome.cnv.rdnacopy as rdnacopy

import handygenome.genomedf.genomedf_utils as genomedf_utils


MERGETYPES_SIMPLEAGG = ['mean', 'std', 'first', 'last', 'max', 'min', 'median', 'skew']
MERGETYPES_WEIGHTED_SIMPLEAGG = ['weighted_mean']
MERGETYPES_BYLENGTH = ['longest', 'shortest']
RESERVED_COLNAMES = [
    'Distance',
    'Overlap',
    'Start_b', 
    'End_b',
]


def split_left_gdf(gdf, width=1000):
    partial_left_gdf_list = list()
    for chrom, subgdf in gdf.group_bychrom(sort=True).items():
        partial_left_gdf_list.extend(subgdf.equal_nrow_split(width=width))

    return partial_left_gdf_list


def targetfunc(partial_left_gdf, right_gdf, join_kwargs):
    subset_args = (
        partial_left_gdf.chroms[0], 
        partial_left_gdf.start0s[0], 
        partial_left_gdf.end0s[-1], 
    )
    partial_right_gdf = right_gdf.subset(*subset_args)
    result = join_base(partial_left_gdf, partial_right_gdf, **join_kwargs)
    return result


###################


@deco.get_deco_arg_choices({'how': [None, 'left']})
def join_base(
    left_gdf, right_gdf, 
    right_gdf_cols=None,
    how='left', 
    merge=['mean', 'std'],
    ddof=0,
    overlapping_length=False,
    N_colname='N',
    suffixes=None,
    merge_args=tuple(),
    merge_kwargs=dict(),
    winsorize=None,
    omit_N=False,
    find_nearest=False,
):
    # merge arghandler
    merge, mergetype_groups = merge_arghandler(merge)
    # suffixes arghandler
    suffixes = suffixes_arghandler(suffixes, merge)
    # right_gdf-related argument handlers
    right_gdf_cols, cols_tobe_added, winsorize = sanitycheck_arghandler(
        left_gdf, right_gdf, right_gdf_cols, merge, N_colname, suffixes, winsorize, omit_N,
    )
        # cols_tobe_added: later will be used for merged_right column ordering

    # handle non-overlap case
    no_overlap = precheck_nonoverlap(left_gdf, right_gdf)
    if no_overlap:
        return make_nonoverlap_result(left_gdf, how, cols_tobe_added)

    # make parameters
    left_columns = left_gdf.columns
    report_overlap = (
        overlapping_length 
        and (merge is not None)
        and set(merge).intersection(MERGETYPES_BYLENGTH)
    )

    # run pyranges join
    left_gr = left_gdf.gr
    right_gr = right_gdf.gr[right_gdf_cols]
    joined_gr = left_gr.join(
        right_gr, 
        how=how, 
        report_overlap=report_overlap,
    )
    if joined_gr.empty:
        return make_nonoverlap_result(left_gdf, how, cols_tobe_added)

    joined_df = joined_gr.df
    joined_df.reset_index(drop=True, inplace=True)

    # handle unmatched rows
    if find_nearest:
        # this function modifies "joined_df" inplace
        add_nearest(joined_df, right_gdf, left_columns)

    # replace missing values with pd.NA
    unmatched_selector = (joined_df['Start_b'] == -1).to_numpy()
    joined_df.loc[
        unmatched_selector, 
        ~joined_df.columns.isin(left_columns)
    ] = pd.NA  # this becomes np.nan in float columns

    # merge
    if merge is None:
        result_df = joined_df
    else:
        # unmatched rows
        unmatched_df = joined_df.loc[unmatched_selector, left_columns]
        if not omit_N:
            unmatched_df.insert(unmatched_df.shape[1], N_colname, 0)

        # matched left
        matched_left = joined_df.loc[~unmatched_selector, left_columns]
        counts, groupkey = genomedf_utils.get_coord_groupkey(matched_left, left_gdf.chromdict)
            # "counts" is aligned with "groupkey"

        merged_left = matched_left.drop_duplicates()
        merged_left.reset_index(drop=True, inplace=True)
        if not omit_N:
            merged_left.insert(merged_left.shape[1], N_colname, counts)

        # matched right
        matched_right = joined_df.loc[
            ~unmatched_selector, 
            ~joined_df.columns.isin(left_columns)
        ]
        matched_right.reset_index(drop=True, inplace=True)
        if winsorize is not None:
            winsorize_matched_right(
                matched_right, 
                groupkey, 
                winsorize,
            )

        merged_right_list = [
            merge_right(
                matched_right, 
                groupkey, 
                right_gdf_cols, 
                mergetype, 
                suffixes, 
                ddof,
                merge_args, 
                merge_kwargs,
            )
            for mergetype in merge
        ]
        merged_right = pd.concat(merged_right_list, axis=1)

        # concat all
        matched = pd.concat([merged_left, merged_right], axis=1)
        result_df = pd.concat([unmatched_df, matched], axis=0)

        # coerce dtype
        if not omit_N:
            result_df = result_df.astype({N_colname: int})

    # return
    result = left_gdf.spawn(result_df)
    result.sort()
    return result


def precheck_nonoverlap(left_gdf, right_gdf):
    return (
        left_gdf.is_empty
        or right_gdf.is_empty
        or (not set(left_gdf.chroms).intersection(right_gdf.chroms))
        or (left_gdf.end0s.max() <= right_gdf.start0s.min())
        or (left_gdf.start0s.min() >= right_gdf.end0s.max())
    )


def make_nonoverlap_result(left_gdf, how, cols_tobe_added):
    if how is None:
        result = left_gdf.iloc[[], :]
    elif how == 'left':
        result = left_gdf.copy()

    for x in cols_tobe_added:
        result[x] = np.nan

    return result


def add_nearest(joined_df, right_gdf, left_columns):
    unmatched_selector = (joined_df['Start_b'] == -1).to_numpy()
    unmatched_df = joined_df.iloc[np.nonzero(unmatched_selector)[0], :3]
    unmatched_df_annot = pr.PyRanges(unmatched_df).nearest(right_gdf.gr).df
    annot_cols = unmatched_df_annot.columns[3:].to_list()  # includes Start_b, End_b, Distance

    joined_df['Distance'] = np.nan  # joined_df already includes columns "Start_b" and "End_b"
    joined_df.loc[unmatched_selector, annot_cols] = unmatched_df_annot.loc[:, annot_cols].to_numpy()


def group_mergetypes(merge):
    mergetype_groups = {
        'simpleagg': list(),
        'weighted_simpleagg': list(),
        'bylength': list(),
        'custom': list(),
    }
    for mergetype in merge:
        if mergetype in MERGETYPES_SIMPLEAGG:
            mergetype_groups['simpleagg'].append(mergetype)
        elif mergetype in MERGETYPES_WEIGHTED_SIMPLEAGG:
            mergetype_groups['weighted_simpleagg'].append(mergetype)
        elif mergetype in MERGETYPES_BYLENGTH:
            mergetype_groups['bylength'].append(mergetype)
        else:
            mergetype_groups['custom'].append(mergetype)

    return mergetype_groups


def merge_arghandler(merge):
    if merge is None:
        mergetype_groups = None
    else:
        # modification
        merge = np.atleast_1d(merge)
        assert len(merge) > 0
        mergetype_groups = group_mergetypes(merge)

        # sanitycheck
        if not all(map(callable, mergetype_groups['custom'])):
            raise Exception(
                f'Custom merge methods must be non-lambda callables'
            )
        if any(map(tools.check_is_lambda, mergetype_groups['custom'])):
            raise Exception(f'Custom merge methods must not include lambda')

    return merge, mergetype_groups


def get_suffixes_key(mergetype):
    if isinstance(mergetype, str):
        return mergetype
    elif callable(mergetype):
        return mergetype.__name__
    else:
        raise Exception(f'Unavailable mergetype')


def suffixes_arghandler(suffixes, merge):
    if merge is None:
        pass
    else:
        # make default suffixes
        default_suffixes_keys = list()
        default_suffixes_vals = list()

        for mergetype in merge:
            key = get_suffixes_key(mergetype)
            if isinstance(mergetype, str):
                val = f'_{mergetype}'
            elif callable(mergetype):
                val = f'_{mergetype.__name__}'

            default_suffixes_keys.append(key)
            default_suffixes_vals.append(val)

        if not tools.check_unique(default_suffixes_keys):
            raise Exception(f'Overlapping merge method names')
        if not tools.check_unique(default_suffixes_vals):
            raise Exception(f'Overlapping merge method suffixes')

        default_suffixes = dict(
            zip(default_suffixes_keys, default_suffixes_vals)
        )

        # apply custom suffixes
        if suffixes is None:
            suffixes = default_suffixes
        else:
            assert isinstance(suffixes, dict)
            suffixes = default_suffixes | suffixes
            if not tools.check_unique(suffixes.values()):
                raise Exception(f'Overlapping merge method suffixes')

    return suffixes


def sanitycheck_arghandler(
    left_gdf, right_gdf, right_gdf_cols, merge, N_colname, suffixes, winsorize, omit_N,
):
    ########################################
    # sanitycheck BEFORE args modification #
    ########################################
    assert len(left_gdf.columns) == len(set(left_gdf.columns))
    assert len(right_gdf.columns) == len(set(right_gdf.columns))

    assert not left_gdf.check_duplicate_coords()
    assert not right_gdf.check_duplicate_coords()

    #####################
    # args modification #
    #####################

    # right_gdf_cols is a list
    if right_gdf_cols is None:
        right_gdf_cols = list(right_gdf.annot_cols)
    else:
        right_gdf_cols = list(np.atleast_1d(right_gdf_cols))

    if winsorize is not None:
        winsorize = tuple(winsorize)
        assert len(winsorize) == 2

    #######################################
    # sanitycheck AFTER args modification #
    #######################################

    assert len(right_gdf_cols) == len(set(right_gdf_cols))
    assert set(right_gdf_cols).issubset(right_gdf.annot_cols)
    assert not set(left_gdf.annot_cols).intersection(right_gdf_cols), (
        f'There must be no overlapping annotation columns between '
        f'two {left_gdf.__class__.__name__} objects.'
    )

    reserved_colnames = list(RESERVED_COLNAMES)
    if not omit_N:
        reserved_colnames.append(N_colname)

    cols_after_pyranges_join = list(left_gdf.columns) + right_gdf_cols
    assert not set(reserved_colnames).intersection(cols_after_pyranges_join)

    if merge is None:
        cols_tobe_added = right_gdf_cols
    else:
        cols_tobe_added = list(
            (x + y) for (x, y) 
            in itertools.product(right_gdf_cols, suffixes.values())
        )

    assert not set(left_gdf.annot_cols).intersection(cols_tobe_added)
    assert not set(reserved_colnames).intersection(cols_tobe_added)

    return right_gdf_cols, cols_tobe_added, winsorize


def winsorize_matched_right(
    matched_right, 
    groupkey, 
    winsorize,
):
    new_values_dict = dict()
    for colname in matched_right.columns:
        new_values = list()
        ser = matched_right[colname]
        for key, val in ser.groupby(groupkey):
            new_val = scipy.stats.mstats.winsorize(
                val, 
                limits=winsorize,
                nan_policy='omit',
            )
            new_values.append(np.ma.getdata(new_val))

        new_values_dict[colname] = np.concatenate(new_values)

    for colname, new_values in new_values_dict.items():
        matched_right[colname] = new_values


def merge_right(
    matched_right, 
    groupkey, 
    right_gdf_cols, 
    mergetype, 
    suffixes, 
    ddof,
    merge_args, 
    merge_kwargs,
):
    if mergetype in MERGETYPES_SIMPLEAGG:
        merged_right = merge_right_simpleagg(
            matched_right, groupkey, right_gdf_cols, mergetype, ddof, 
        )
    elif mergetype in MERGETYPES_WEIGHTED_SIMPLEAGG:
        merged_right = merge_right_weighted_simpleagg(
            matched_right, groupkey, right_gdf_cols, mergetype, ddof,
        )
    elif mergetype in MERGETYPES_BYLENGTH:
        merged_right = merge_right_bylength(
            matched_right, groupkey, right_gdf_cols, mergetype, ddof,
        )
    else:  # custom merge method
        merged_right = merge_right_custom(
            matched_right, 
            groupkey, 
            right_gdf_cols, 
            mergetype, 
            merge_args, 
            merge_kwargs,
        )

    merged_right.reset_index(drop=True, inplace=True)
    suffix = suffixes[get_suffixes_key(mergetype)]
    merged_right = merged_right.add_suffix(suffix, axis=1)

    return merged_right


def merge_right_custom(
    matched_right, 
    groupkey, 
    right_gdf_cols, 
    mergetype, 
    merge_args, 
    merge_kwargs,
):
    '''When *args or **kwargs are given to "agg", "func" receives
    DataFrame; otherwise "func" receives Series.
    '''
    aggresult_list = list()
    for x in right_gdf_cols:
        ser = matched_right.loc[:, x]
        aggresult = ser.groupby(groupkey, sort=False).agg(
            mergetype, 
            *merge_args, 
            engine=None, 
            engine_kwargs=None,
            **merge_kwargs,
        )
        aggresult.reset_index(drop=True, inplace=True)
        aggresult_list.append(aggresult)

    merged_right = pd.concat(aggresult_list, axis=1)
    return merged_right


def merge_right_simpleagg(
    matched_right, groupkey, right_gdf_cols, mergetype, ddof,
):
    assert mergetype in MERGETYPES_SIMPLEAGG

    groupby_obj = matched_right.groupby(groupkey, sort=False)[right_gdf_cols]

    if mergetype == 'first':
        merged_right = groupby_obj.first()
    elif mergetype == 'last':
        merged_right = groupby_obj.last()
    elif mergetype == 'mean':
        merged_right = groupby_obj.mean()
    elif mergetype == 'std':
        merged_right = groupby_obj.std(ddof=ddof)
    elif mergetype == 'max':
        merged_right = groupby_obj.max()
    elif mergetype == 'min':
        merged_right = groupby_obj.min()
    elif mergetype == 'median':
        merged_right = groupby_obj.median()
    elif mergetype == 'skew':
        merged_right = groupby_obj.skew()
    else:
        raise Exception(f'Unknown mergetype: {mergetype}')

    return merged_right


def merge_right_weighted_simpleagg(matched_right, groupkey, right_gdf_cols, mergetype, ddof):
    assert mergetype in MERGETYPES_WEIGHTED_SIMPLEAGG
    #assert (matched_right.index == pd.RangeIndex(start=0, end=matched_right.shape[0])).all()

    new_matched_right = make_weighted_right(matched_right, right_gdf_cols)
    groupby_obj = new_matched_right.groupby(groupkey, sort=False)[right_gdf_cols]

    if mergetype == 'weighted_mean':
        merged_right = groupby_obj.sum()
    else:
        raise Exception(f'Unknown mergetype: {mergetype}')

    return merged_right


def make_weighted_right(matched_right, right_gdf_cols):
    if 'Overlap' in matched_right.columns:
        lengths = matched_right['Overlap']
    else:
        lengths = matched_right['End_b'] - matched_right['Start_b']

    lengthsum = lengths.sum()

    src_data = {
        key: ((matched_right[key] * lengths) / lengthsum)
        for key in right_gdf_cols
    }
    weighted_matched_right = pd.DataFrame(src_data)

    return weighted_matched_right


def merge_right_bylength(matched_right, groupkey, right_gdf_cols, mergetype, ddof):
    assert mergetype in MERGETYPES_BYLENGTH
    #assert (matched_right.index == pd.RangeIndex(start=0, end=matched_right.shape[0])).all()

    if 'Overlap' in matched_right.columns:
        lengths = matched_right['Overlap']
    else:
        lengths = matched_right['End_b'] - matched_right['Start_b']

    groupby_obj = lengths.reset_index(drop=True).groupby(groupkey, sort=False)
    if mergetype == 'longest':
        indexes = groupby_obj.idxmax()
    elif mergetype == 'shortest':
        indexes = groupby_obj.idxmin()
    else:
        raise Exception(f'Unknown mergetype: {mergetype}')

    right_gdf_cols_idxs = [matched_right.columns.get_loc(x) for x in right_gdf_cols]
    merged_right = matched_right.iloc[indexes.to_numpy(), right_gdf_cols_idxs]

    return merged_right


########
# isec #
########

def isec_base(left_gdf, right_gdf):
    """Assume left_gdf and right_gdf has single identical chromosome"""

    left_gdf.sort()
    right_gdf.sort()

    left_start0s = left_gdf.start0s
    left_end0s = left_gdf.end0s
    right_start0s = right_gdf.start0s
    right_end0s = right_gdf.end0s


# isec boolean array makers

def check_intersect_num_leading_true(arr):
    result = tools.array_index(arr, False)
    if result is None:
        result = arr.shape[0]
    return result


def check_intersect_num_leading_false(arr):
    result = tools.array_index(arr, True)
    if result is None:
        result = arr.shape[0]
    return result


def check_intersect(
    left_start0s, left_end0s, right_start0s, right_end0s, 
    no_inner_overlap=False,
    sanitycheck=True,
):
    if (no_inner_overlap and sanitycheck):
        assert not genomedf_utils.check_interval_overlap(left_start0s, left_end0s)
        assert not genomedf_utils.check_interval_overlap(right_start0s, right_end0s)

    left_start0s = np.asarray(left_start0s)
    left_end0s = np.asarray(left_end0s)
    right_start0s = np.asarray(right_start0s)
    right_end0s = np.asarray(right_end0s)

    nonovlp_selector = np.logical_or(
        (left_end0s <= right_start0s[0]),
        (left_start0s >= right_end0s[-1]),
    )

    result = np.empty((len(left_start0s), len(right_start0s)), dtype=bool)
    result[nonovlp_selector, :] = False

    main_func = (
        check_intersect_relevant_noinneroverlap
        if no_inner_overlap else
        check_intersect_relevant
    )
    result[~nonovlp_selector, :] = main_func(
        left_start0s[~nonovlp_selector],
        left_end0s[~nonovlp_selector],
        right_start0s,
        right_end0s,
    )

    return result
    

def check_intersect_relevant(left_start0s, left_end0s, right_start0s, right_end0s):
    # less than left end
    rstart_lt_lend_list = list()
    for end0 in left_end0s:
        current_selector = (right_start0s < end0)
        rstart_lt_lend_list.append(current_selector)

    rstart_lt_lend = np.stack(rstart_lt_lend_list, axis=0)

    # greater than left start
    for lt_lend_selector, start0 in zip(rstart_lt_lend, left_start0s):
        lt_lend_selector[lt_lend_selector] = (
            right_end0s[lt_lend_selector] > start0
        )

    return rstart_lt_lend


def check_intersect_relevant_noinneroverlap(left_start0s, left_end0s, right_start0s, right_end0s):
    # less than left end
    rstart_lt_lend_list = list()
    numtrue_list = list()
    for end0 in left_end0s:
        if len(numtrue_list) == 0:
            last_numtrue = 0
        else:
            last_numtrue = numtrue_list[-1]

        current_selector = np.concatenate(
            [
                np.repeat(True, last_numtrue),
                (right_start0s[last_numtrue:] < end0),
            ]
        )
        rstart_lt_lend_list.append(current_selector)
        numtrue_list.append(check_intersect_num_leading_true(current_selector))

    rstart_lt_lend = np.stack(rstart_lt_lend_list, axis=0)

    #print(rstart_lt_lend)

    # greater than left start
    #last_num_leadingfalse_list = list()
    #last_leadingfalse = 0
    last_idx = len(left_start0s) - 1
    for idx, (start0, numtrue) in enumerate(zip(left_start0s, numtrue_list)):
        #if len(last_num_leadingfalse_list) == 0:
        #    last_leadingfalse = 0
        #else:
        #    last_leadingfalse = last_num_leadingfalse_list[-1]
        #    last_leadingfalse = check_intersect_num_leading_false(rstart_lt_lend[idx - 1, :])

        subarr = rstart_lt_lend[idx, :]
        subarr[subarr] = right_end0s[subarr] > start0
        if idx != last_idx:
            rstart_lt_lend[idx + 1, :numtrue] = subarr[:numtrue]

        #last_leadingfalse = check_intersect_num_leading_false(subsubarr)

    return rstart_lt_lend


